@@ -11,6 +11,11 @@
 
 #define	MAXTIMESTR	8
 #define	ATTRWIDTH	10
+#if	MSDOS
+#define	DIRENTSIZ(s)	DOSDIRENT
+#else
+#define	DIRENTSIZ(s)	(sizeof(struct dirent) - DNAMESIZE + strlen(s) + 1)
+#endif
 
 typedef struct _attrib_t {
 	short nlink;
@@ -36,7 +41,7 @@ extern int lcmdline;
 extern int maxcmdline;
 extern int mark;
 #ifdef	HAVEFLAGS
-extern u_long fflaglist[];
+extern CONST u_long fflaglist[];
 #endif
 #if	!defined (_USEDOSCOPY) && !defined (_NOEXTRACOPY)
 extern int inheritcopy;
@@ -54,26 +59,33 @@ extern int inheritcopy;
 #define	FLAG_SAMEDIR	8
 #endif
 
-static int NEAR issamedir __P_((char *, char *));
+static int NEAR issamedir __P_((CONST char *, CONST char *));
 static int NEAR islowerdir __P_((VOID_A));
-static char *NEAR getdestdir __P_((char *, char *));
-static int NEAR getdestpath __P_((char *, char *, struct stat *));
-static int NEAR getcopypolicy __P_((char *));
-static int NEAR getremovepolicy __P_((char *));
-static int NEAR checkdupl __P_((char *, char *, struct stat *, struct stat *));
-static int NEAR checkrmv __P_((char *, int));
+static char *NEAR getdestdir __P_((CONST char *, CONST char *));
+static int NEAR getdestpath __P_((CONST char *, char *, struct stat *));
+static int NEAR getcopypolicy __P_((CONST char *));
+static int NEAR getremovepolicy __P_((CONST char *, int));
+static int NEAR checkdupl __P_((CONST char *, char *,
+		struct stat *, struct stat *));
+static int NEAR checkrmv __P_((CONST char *, int));
 #ifndef	_NOEXTRACOPY
-static int NEAR isxdev __P_((char *, struct stat *));
+static int NEAR isxdev __P_((CONST char *, struct stat *));
+static VOID NEAR addcopysize __P_((off_t));
+static VOID NEAR _showprogress __P_((int));
+static int countcopysize __P_((CONST char *));
+static int countmovesize __P_((CONST char *));
+static int NEAR prepareprogress __P_((int, int (*)(CONST char *)));
 #endif
-static VOID NEAR preparecopy __P_((VOID_A));
-static VOID NEAR preparemove __P_((VOID_A));
-static int safecopy __P_((char *));
-static int safemove __P_((char *));
-static int cpdir __P_((char *));
-static int touchdir __P_((char *));
+static int NEAR preparecopy __P_((int, int));
+static int NEAR preparemove __P_((int, int));
+static int safecopy __P_((CONST char *));
+static int safemove __P_((CONST char *));
+static int cpdir __P_((CONST char *));
+static int touchdir __P_((CONST char *));
 #ifndef	_NOEXTRACOPY
-static int mvdir1 __P_((char *));
-static int mvdir2 __P_((char *));
+static int mvdir1 __P_((CONST char *));
+static int mvdir2 __P_((CONST char *));
+static int countremovesize __P_((CONST char *));
 #endif
 static VOID NEAR showmode __P_((attrib_t *, int, int));
 static VOID NEAR showattr __P_((namelist *, attrib_t *, int));
@@ -82,13 +94,18 @@ static int NEAR inputuid __P_((attrib_t *, int));
 static int NEAR inputgid __P_((attrib_t *, int));
 #endif
 static char **NEAR getdirtree __P_((char *, char **, int *, int));
-static int NEAR _applydir __P_((char *, int (*)(char *),
-		int (*)(char *), int (*)(char *), int, char *, int));
-static int forcecpfile __P_((char *));
-static int forcecpdir __P_((char *));
-static int forcetouchdir __P_((char *));
+static int NEAR _applydir __P_((CONST char *, int (*)(CONST char *),
+		int (*)(CONST char *), int (*)(CONST char *),
+		int, CONST char *, int));
+static int forcecpfile __P_((CONST char *));
+static int forcecpdir __P_((CONST char *));
+static int forcetouchdir __P_((CONST char *));
 
 char *destpath = NULL;
+#ifndef	_NOEXTRACOPY
+int progressbar = 0;
+int precopymenu = 0;
+#endif
 
 static int copypolicy = 0;
 static int removepolicy = 0;
@@ -119,37 +136,30 @@ static int destdrive = -1;
 #if	!defined (_NOEXTRACOPY) && !defined (_NODOSDRIVE)
 static int forwarddrive = -1;
 #endif
+#ifndef	_NOEXTRACOPY
+static int copycolumn = -1;
+static int copyline = 0;
+static long copyunit = 0L;
+static long maxcopyunit = 0L;
+static off_t copysize = (off_t)0;
+static off_t maxcopysize = (off_t)0;
+#endif
 
 
 static int NEAR issamedir(path, org)
-char *path, *org;
+CONST char *path, *org;
 {
-	char *cwd;
-	int i;
-
-	cwd = getwd2();
-	if (!org) org = cwd;
-	else if (_chdir2(org) >= 0) {
-		org = getwd2();
-		if (_chdir2(cwd) < 0) error(cwd);
-	}
-	else {
-		free(cwd);
-		return(0);
-	}
+	char path2[MAXPATHLEN], org2[MAXPATHLEN];
 
+	org = (org) ? getrealpath(org, org2, NULL) : Xgetwd(org2);
+	if (org != org2) return(0);
 #ifdef	_USEDOSPATH
-	if (dospath(path, NULL) != dospath(org, NULL)) path = NULL;
-	else
+	if (dospath(path, NULL) != dospath(org2, NULL)) return(0);
 #endif
-	path = (_chdir2(path) >= 0) ? getwd2() : NULL;
-	if (_chdir2(cwd) < 0) error(cwd);
-	i = (path) ? !strpathcmp(org, path) : 0;
-	if (org != cwd) free(org);
-	free(cwd);
-	if (path) free(path);
+	if (getrealpath(path, path2, NULL) != path2) return(0);
+	if (strpathcmp(path2, org2)) return(0);
 
-	return(i);
+	return(1);
 }
 
 static int NEAR islowerdir(VOID_A)
@@ -157,58 +167,41 @@ static int NEAR islowerdir(VOID_A)
 #ifdef	_USEDOSEMU
 	char orgpath[MAXPATHLEN];
 #endif
-	char *cp, *top, *cwd, *path, *org, tmp[MAXPATHLEN];
-	int i;
+	CONST char *org;
+	char *cp, *top, path[MAXPATHLEN], buf[MAXPATHLEN];
+	int len;
 
-	cwd = getwd2();
-	path = destpath;
-	org = filelist[filepos].name;
-	if (!org) org = cwd;
-#ifdef	_USEDOSEMU
-	else org = nodospath(orgpath, org);
-#endif
-	strcpy(tmp, path);
-	top = tmp;
-#ifdef	_USEDOSPATH
-	if (dospath(path, NULL) != dospath(org, NULL)) {
-		free(cwd);
+	if (!isdir(&(filelist[filepos])) || islink(&(filelist[filepos])))
 		return(0);
-	}
-	if (_dospath(tmp)) top += 2;
+	org = filelist[filepos].name;
+#ifdef	_USEDOSEMU
+	org = nodospath(orgpath, org);
+#endif
+	strcpy(path, destpath);
+	top = path;
+#ifdef	_USEDOSPATH
+	if (dospath(path, NULL) != dospath(org, NULL)) return(0);
+	if (_dospath(path)) top += 2;
 #endif
-	while (_chdir2(tmp) < 0) {
-		if ((cp = strrdelim(top, 0)) && cp > top) *cp = '\0';
-		else {
-			path = NULL;
-			break;
-		}
-	}
-	i = 0;
-	if (path) {
-		path = getwd2();
-		if (_chdir2(cwd) < 0) error(cwd);
-		if (_chdir2(org) >= 0) {
-			org = getwd2();
-			i = strlen(org);
-			if (strnpathcmp(org, path, i) || path[i] != _SC_)
-				i = 0;
-			free(org);
-		}
-		if (_chdir2(cwd) < 0) error(cwd);
-		free(path);
-	}
-	free(cwd);
 
-	return(i);
+	while ((cp = (char *)getrealpath(path, buf, NULL)) != buf) {
+		if (cp != path) return(0);
+		cp = strrdelim(top, 0);
+		if (!cp || cp <= top) return(0);
+		*cp = '\0';
+	}
+
+	if (getrealpath(org, path, NULL) != path) return(0);
+	len = strlen(path);
+	if (strnpathcmp(path, buf, len) || buf[len] != _SC_) len = 0;
+
+	return(len);
 }
 
 static char *NEAR getdestdir(mes, arg)
-char *mes, *arg;
+CONST char *mes, *arg;
 {
 	char *dir;
-#ifndef	_NODOSDRIVE
-	int drive;
-#endif
 
 	if (arg && *arg) dir = strdup2(arg);
 	else if (!(dir = inputstr(mes, 1, -1, NULL, HST_PATH))) return(NULL);
@@ -226,9 +219,7 @@ char *mes, *arg;
 #endif
 
 #ifndef	_NODOSDRIVE
-	destdrive = -1;
-	if ((drive = dospath3(dir))
-	&& (destdrive = preparedrv(drive)) < 0) {
+	if (preparedrv(dospath3(dir), &destdrive) < 0) {
 		warning(-1, dir);
 		free(dir);
 		return(NULL);
@@ -238,7 +229,8 @@ char *mes, *arg;
 		warning(-1, dir);
 		free(dir);
 #ifndef	_NODOSDRIVE
-		if (destdrive >= 0) shutdrv(destdrive);
+		shutdrv(destdrive);
+		destdrive = -1;
 #endif
 		return(NULL);
 	}
@@ -247,10 +239,12 @@ char *mes, *arg;
 }
 
 static int NEAR getdestpath(file, dest, stp)
-char *file, *dest;
+CONST char *file;
+char *dest;
 struct stat *stp;
 {
-	char *cp, *tmp;
+	CONST char *cp;
+	char *tmp;
 	int n;
 
 	strcpy(dest, destpath);
@@ -273,12 +267,15 @@ struct stat *stp;
 }
 
 static int NEAR getcopypolicy(s)
-char *s;
+CONST char *s;
 {
-#if	!defined (_NOEXTRACOPY) && !defined (_NODOSDRIVE)
+#ifndef	_NOEXTRACOPY
+# ifndef	_NODOSDRIVE
 	int dupdestdrive;
-#endif
-	char *cp, *str[MAXCOPYITEM];
+# endif
+	char *cp;
+#endif	/* !_NOEXTRACOPY */
+	CONST char *str[MAXCOPYITEM];
 	int n, ch, val[MAXCOPYITEM];
 
 	for (;;) {
@@ -312,6 +309,7 @@ char *s;
 			if (selectstr(&n, 2, 0, str, val) != K_CR) continue;
 # ifndef	_NODOSDRIVE
 			dupdestdrive = destdrive;
+			destdrive = -1;
 # endif
 
 			if (!(cp = getdestdir(FRWDD_K, NULL))) continue;
@@ -335,26 +333,34 @@ char *s;
 	return(copypolicy);
 }
 
-static int NEAR getremovepolicy(s)
-char *s;
+static int NEAR getremovepolicy(s, pre)
+CONST char *s;
+int pre;
 {
-	char *str[4];
-	int n, ch, val[4];
+	CONST char *str[4];
+	int i, n, ch, val[4];
 
 	Xlocate(0, L_CMDLINE);
 	Xputterm(L_CLEAR);
 	Xkanjiputs(s);
 
-	str[0] = ANYES_K;
-	val[0] = CHK_OK;
-	str[1] = ANNO_K;
-	val[1] = CHK_ERROR;
-	str[2] = ANALL_K;
-	val[2] = RMP_REMOVEALL;
-	str[3] = ANKEP_K;
-	val[3] = RMP_KEEPALL;
+	i = 0;
+	if (pre) {
+		str[i] = ANCNF_K;
+		val[i++] = CHK_OK;
+	}
+	else {
+		str[i] = ANYES_K;
+		val[i++] = CHK_OK;
+		str[i] = ANNO_K;
+		val[i++] = CHK_ERROR;
+	}
+	str[i] = ANALL_K;
+	val[i++] = RMP_REMOVEALL;
+	str[i] = ANKEP_K;
+	val[i++] = RMP_KEEPALL;
 	n = CHK_OK;
-	ch = selectstr(&n, 4, 0, str, val);
+	ch = selectstr(&n, i, 0, str, val);
 	if (ch == K_ESC) return(CHK_ERROR);
 	else if (ch != K_CR) return(CHK_CANCEL);
 	else if (n <= 0) return(n);
@@ -364,12 +370,14 @@ char *s;
 }
 
 static int NEAR checkdupl(file, dest, stp1, stp2)
-char *file, *dest;
+CONST char *file;
+char *dest;
 struct stat *stp1, *stp2;
 {
 #ifndef	_NOEXTRACOPY
 	char path[MAXPATHLEN];
 #endif
+	CONST char *s;
 	char *cp, *tmp;
 	int i, n;
 
@@ -394,9 +402,13 @@ struct stat *stp1, *stp2;
 
 	for (;;) {
 		if (!n || n == CPP_RENAME) {
-			cp = SAMEF_K;
-			i = strlen2(cp) - strsize("%.*s");
-			cp = asprintf3(cp, n_column - i, dest);
+			s = SAMEF_K;
+			i = strlen2(s) - strsize("%.*s");
+			cp = asprintf3(s, n_column - i, dest);
+
+#ifndef	_NOEXTRACOPY
+			copycolumn = -1;
+#endif
 			if (!n) n = getcopypolicy(cp);
 			else {
 				Xlocate(0, L_CMDLINE);
@@ -468,18 +480,21 @@ struct stat *stp1, *stp2;
 }
 
 static int NEAR checkrmv(path, mode)
-char *path;
+CONST char *path;
 int mode;
 {
-	char *cp;
-	int n, len;
-#if	MSDOS
-
-	if (Xaccess(path, mode) >= 0) return(CHK_OK);
-#else	/* !MSDOS */
+#if	!MSDOS
 	struct stat *stp, st;
 	char *tmp, dir[MAXPATHLEN];
+	int duperrno;
+#endif
+	CONST char *s;
+	char *cp;
+	int n, len;
 
+#if	MSDOS
+	if (Xaccess(path, mode) >= 0) return(CHK_OK);
+#else	/* !MSDOS */
 # ifndef	_NODOSDRIVE
 	if (dospath2(path)) {
 		if (Xaccess(path, mode) >= 0) return(CHK_OK);
@@ -515,11 +530,9 @@ int mode;
 
 #if	!MSDOS
 	if (stp) {
-		int duperrno;
-
 		duperrno = errno;
 # ifdef	NOUID
-		mode = logical_access(stp -> st_mode)
+		mode = logical_access(stp -> st_mode);
 # else
 		mode = logical_access(stp -> st_mode,
 			stp -> st_uid, stp -> st_gid);
@@ -534,10 +547,13 @@ int mode;
 
 	if (removepolicy > 0) n = removepolicy;
 	else {
-		cp = DELPM_K;
-		len = strlen2(cp) - strsize("%.*s");
-		cp = asprintf3(cp, n_column - len, path);
-		n = getremovepolicy(cp);
+		s = DELPM_K;
+		len = strlen2(s) - strsize("%.*s");
+		cp = asprintf3(s, n_column - len, path);
+#ifndef	_NOEXTRACOPY
+		copycolumn = -1;
+#endif
+		n = getremovepolicy(cp, 0);
 		free(cp);
 	}
 	if (n > 0) n -= RMP_BIAS;
@@ -548,7 +564,7 @@ int mode;
 #ifndef	_NOEXTRACOPY
 /*ARGSUSED*/
 static int NEAR isxdev(path, stp)
-char *path;
+CONST char *path;
 struct stat *stp;
 {
 # if	!MSDOS
@@ -570,24 +586,168 @@ struct stat *stp;
 
 	return(0);
 }
+
+static VOID NEAR addcopysize(size)
+off_t size;
+{
+	if (size <= MAXTYPE(off_t) - maxcopysize) maxcopysize += size;
+	else {
+		maxcopysize -= MAXTYPE(off_t) - size;
+		maxcopyunit++;
+	}
+}
+
+static VOID NEAR _showprogress(n)
+int n;
+{
+	int i, max;
+
+	max = n_column - 1;
+	Xlocate(0, copyline);
+	for (i = 0; i < n; i++) Xputch2('o');
+	for (; i < max; i++) Xputch2('.');
+	Xlocate(n, copyline);
+	Xtflush();
+}
+
+VOID showprogress(size)
+off_t size;
+{
+	long unit;
+	int n, max;
+
+	if (!maxcopysize && !maxcopyunit) return;
+
+	max = n_column - 1;
+	for (n = 0; n < max; n++) {
+		if (size <= MAXTYPE(off_t) - copysize) copysize += size;
+		else {
+			copysize -= MAXTYPE(off_t) - size + 1;
+			copyunit++;
+		}
+	}
+
+	size = copysize;
+	unit = copyunit;
+	for (n = 0; n < max; n++) {
+		if (maxcopysize <= size) size -= maxcopysize;
+		else if (!unit) break;
+		else {
+			size += MAXTYPE(off_t) - maxcopysize + 1;
+			unit--;
+		}
+		if (unit < maxcopyunit) break;
+		unit -= maxcopyunit;
+	}
+
+	if (n > copycolumn) _showprogress(copycolumn = n);
+}
+
+VOID fshowprogress(path)
+CONST char *path;
+{
+	CONST char *cp;
+
+	if (!(cp = strrdelim(path, 1))) cp = path;
+	showprogress(DIRENTSIZ(cp));
+}
+
+static int countcopysize(path)
+CONST char *path;
+{
+	struct stat st;
+	CONST char *cp;
+
+	if (!(cp = strrdelim(path, 1))) cp = path;
+	addcopysize(DIRENTSIZ(cp));
+	if (Xlstat(path, &st) < 0) return(APL_ERROR);
+	addcopysize(st.st_size);
+
+	return(APL_OK);
+}
+
+static int countmovesize(path)
+CONST char *path;
+{
+	struct stat st;
+	int n;
+	CONST char *cp;
+
+	if (!(cp = strrdelim(path, 1))) cp = path;
+	addcopysize(DIRENTSIZ(cp));
+	if ((n = isxdev(path, &st)) < 0) return(APL_ERROR);
+	else if (n && !s_isdir(&st)) addcopysize(st.st_size);
+
+	return(APL_OK);
+}
+
+static int NEAR prepareprogress(isdir, func)
+int isdir;
+int (*func)__P_((CONST char *));
+{
+	int ret;
+
+	copycolumn = -1;
+	copyline = 0;
+	copyunit = maxcopyunit = 0L;
+	copysize = maxcopysize = (off_t)0;
+	if (progressbar) {
+		if (!isdir) {
+			copyline = L_CMDLINE;
+			VOID_C applyfile(func, NULL);
+		}
+		else {
+			copyline = L_STACK;
+			ret = applydir(filelist[filepos].name, func,
+				NULL, func, ORD_NOPREDIR, NULL);
+			if (ret == APL_ERROR) return(-1);
+			if (ret == APL_CANCEL) {
+				maxcopyunit = 0L;
+				maxcopysize = (off_t)0;
+			}
+		}
+	}
+
+	return(0);
+}
 #endif	/* !_NOEXTRACOPY */
 
-static VOID NEAR preparecopy(VOID_A)
+/*ARGSUSED*/
+static int NEAR preparecopy(isdir, narg)
+int isdir, narg;
 {
 	copypolicy = (issamedir(destpath, NULL))
 		? (FLAG_SAMEDIR | CPP_RENAME) : 0;
+#ifndef	_NOEXTRACOPY
+	if (precopymenu && !copypolicy && (isdir || narg > 1))
+		if (getcopypolicy(PRECP_K) < 0) return(-1);
+	if (prepareprogress(isdir, countcopysize)) return(-1);
+#endif	/* !_NOEXTRACOPY */
 #ifndef	_NODOSDRIVE
 	if (dospath3(nullstr)) waitmes();
 #endif
+
+	return(0);
 }
 
-static VOID NEAR preparemove(VOID_A)
+/*ARGSUSED*/
+static int NEAR preparemove(isdir, narg)
+int isdir, narg;
 {
 	copypolicy = removepolicy = 0;
+#ifndef	_NOEXTRACOPY
+	if (precopymenu && (isdir || narg > 1)) {
+		if (getcopypolicy(PRECP_K) < 0) return(-1);
+		if (getremovepolicy(PRERM_K, 1) < 0) return(-1);
+	}
+	if (prepareprogress(isdir, countmovesize) < 0) return(-1);
+#endif	/* !_NOEXTRACOPY */
+
+	return(0);
 }
 
 static int safecopy(path)
-char *path;
+CONST char *path;
 {
 	struct stat st1, st2;
 	char dest[MAXPATHLEN];
@@ -601,7 +761,7 @@ char *path;
 }
 
 static int safemove(path)
-char *path;
+CONST char *path;
 {
 	struct stat st1, st2;
 	char dest[MAXPATHLEN];
@@ -616,7 +776,7 @@ char *path;
 }
 
 static int cpdir(path)
-char *path;
+CONST char *path;
 {
 	struct stat st1, st2;
 	char dest[MAXPATHLEN];
@@ -667,13 +827,13 @@ char *path;
 
 /*ARGSUSED*/
 static int touchdir(path)
-char *path;
+CONST char *path;
 {
 	struct stat st;
 	char dest[MAXPATHLEN];
 
 	if (getdestpath(path, dest, &st) < 0) return(APL_CANCEL);
-	st.st_nlink = destnlink;
+	st.st_nlink = (destnlink | TCH_IGNOREERR);
 	st.st_mode = destmode;
 	st.st_mtime = destmtime;
 	st.st_atime = destatime;
@@ -691,7 +851,7 @@ char *path;
 
 #ifndef	_NOEXTRACOPY
 static int mvdir1(path)
-char *path;
+CONST char *path;
 {
 	struct stat st;
 	int n;
@@ -712,54 +872,83 @@ char *path;
 }
 
 static int mvdir2(path)
-char *path;
+CONST char *path;
 {
 	struct stat st;
 	char dest[MAXPATHLEN];
 
 	if (getdestpath(path, dest, &st) < 0) return(APL_CANCEL);
-	st.st_nlink = destnlink;
+	st.st_nlink = (destnlink | TCH_IGNOREERR);
 	st.st_mode = destmode;
 	st.st_mtime = destmtime;
 	st.st_atime = destatime;
 	if (touchfile(dest, &st) < 0) return(APL_ERROR);
 	if (Xrmdir(path) < 0) return(APL_ERROR);
+#ifndef	_NOEXTRACOPY
+	fshowprogress(path);
+#endif
+
+	return(APL_OK);
+}
+
+static int countremovesize(path)
+CONST char *path;
+{
+	CONST char *cp;
+
+	if (!(cp = strrdelim(path, 1))) cp = path;
+	addcopysize(DIRENTSIZ(cp));
 
 	return(APL_OK);
 }
 #endif	/* !_NOEXTRACOPY */
 
-VOID prepareremove(VOID_A)
+/*ARGSUSED*/
+int prepareremove(isdir, narg)
+int isdir, narg;
 {
 	removepolicy = 0;
+#ifndef	_NOEXTRACOPY
+	if (precopymenu && (isdir || narg > 1))
+		if (getremovepolicy(PRERM_K, 1) < 0) return(-1);
+	if (prepareprogress(isdir, countremovesize)) return(-1);
+#endif	/* !_NOEXTRACOPY */
+
+	return(0);
 }
 
 int rmvfile(path)
-char *path;
+CONST char *path;
 {
 	int n;
 
 	if ((n = checkrmv(path, W_OK)) < 0)
 		return((n == CHK_CANCEL) ? APL_CANCEL : APL_IGNORE);
 	if (Xunlink(path) < 0) return(APL_ERROR);
+#ifndef	_NOEXTRACOPY
+	fshowprogress(path);
+#endif
 
 	return(APL_OK);
 }
 
 int rmvdir(path)
-char *path;
+CONST char *path;
 {
 	int n;
 
 	if ((n = checkrmv(path, R_OK | W_OK | X_OK)) < 0)
 		return((n == CHK_CANCEL) ? APL_CANCEL : APL_IGNORE);
 	if (Xrmdir(path) < 0) return(APL_ERROR);
+#ifndef	_NOEXTRACOPY
+	fshowprogress(path);
+#endif
 
 	return(APL_OK);
 }
 
 int findfile(path)
-char *path;
+CONST char *path;
 {
 	if (regexp_exec(findregexp, getbasename(path), 1)) {
 		if (path[0] == '.' && path[1] == _SC_) path += 2;
@@ -776,7 +965,7 @@ char *path;
 }
 
 int finddir(path)
-char *path;
+CONST char *path;
 {
 	if (regexp_exec(findregexp, getbasename(path), 1)) {
 		if (yesno(FOUND_K)) {
@@ -1380,7 +1569,7 @@ int flag;
 }
 
 int setattr(path)
-char *path;
+CONST char *path;
 {
 	struct stat st;
 
@@ -1403,8 +1592,8 @@ char *path;
 }
 
 int applyfile(func, endmes)
-int (*func)__P_((char *));
-char *endmes;
+int (*func)__P_((CONST char *));
+CONST char *endmes;
 {
 #ifdef	_USEDOSEMU
 	char path[MAXPATHLEN];
@@ -1418,12 +1607,15 @@ char *endmes;
 		i = (*func)(fnodospath(path, filepos));
 		if (i == APL_ERROR) warning(-1, filelist[filepos].name);
 		else if (i == APL_OK) ret++;
+#ifndef	_NOEXTRACOPY
+		if (endmes && copyline > 0) _showprogress(n_column);
+#endif
 		return(ret);
 	}
 
 	helpbar();
 	for (filepos = 0; filepos < maxfile; filepos++) {
-		if (intrkey()) {
+		if (intrkey(K_ESC)) {
 			endmes = NULL;
 			break;
 		}
@@ -1443,7 +1635,12 @@ char *endmes;
 		old = filepos;
 	}
 
-	if (endmes) warning(0, endmes);
+	if (endmes) {
+#ifndef	_NOEXTRACOPY
+		if (copyline > 0) _showprogress(n_column);
+#endif
+		warning(0, endmes);
+	}
 	filepos = dupfilepos;
 	movepos(old, 0);
 	Xlocate(win_x, win_y);
@@ -1483,12 +1680,12 @@ int *maxp, depth;
 }
 
 static int NEAR _applydir(dir, funcf, funcd1, funcd2, order, endmes, verbose)
-char *dir;
-int (*funcf)__P_((char *));
-int (*funcd1)__P_((char *));
-int (*funcd2)__P_((char *));
+CONST char *dir;
+int (*funcf)__P_((CONST char *));
+int (*funcd1)__P_((CONST char *));
+int (*funcd2)__P_((CONST char *));
 int order;
-char *endmes;
+CONST char *endmes;
 int verbose;
 {
 	DIR *dirp;
@@ -1496,15 +1693,16 @@ int verbose;
 	struct stat st;
 	time_t dupmtime, dupatime;
 	u_short dupmode;
-	char *cp, *fname, path[MAXPATHLEN], **dirlist;
+	CONST char *cp;
+	char *fname, path[MAXPATHLEN], **dirlist;
 	int ret, ndir, max, dupnlink;
 
-	if (intrkey()) return(APL_CANCEL);
+	if (intrkey(K_ESC)) return(APL_CANCEL);
 
 	if (verbose) {
 		Xlocate(0, L_CMDLINE);
 		Xputterm(L_CLEAR);
-		cp = (dir[0] == '.' && dir[1] == _SC_) ? dir + 2 : dir;
+		cp = (dir[0] == '.' && dir[1] == _SC_) ? &(dir[2]) : dir;
 		Xcprintf2("[%.*k]", n_column - 2, cp);
 		Xtflush();
 	}
@@ -1566,7 +1764,7 @@ int verbose;
 	if (!(dirp = Xopendir(dir))) warning(-1, dir);
 	else {
 		while ((dp = Xreaddir(dirp))) {
-			if (intrkey()) {
+			if (intrkey(K_ESC)) {
 				ret = APL_CANCEL;
 				break;
 			}
@@ -1592,18 +1790,23 @@ int verbose;
 		if (ret == APL_CANCEL) return(ret);
 		warning(-1, dir);
 	}
-	if (endmes) warning(0, endmes);
+	if (endmes) {
+#ifndef	_NOEXTRACOPY
+		if (copyline > 0) _showprogress(n_column);
+#endif
+		warning(0, endmes);
+	}
 
 	return(ret);
 }
 
 int applydir(dir, funcf, funcd1, funcd2, order, endmes)
-char *dir;
-int (*funcf)__P_((char *));
-int (*funcd1)__P_((char *));
-int (*funcd2)__P_((char *));
+CONST char *dir;
+int (*funcf)__P_((CONST char *));
+int (*funcd1)__P_((CONST char *));
+int (*funcd2)__P_((CONST char *));
 int order;
-char *endmes;
+CONST char *endmes;
 {
 	char path[MAXPATHLEN];
 	int verbose;
@@ -1626,7 +1829,7 @@ char *endmes;
 
 /*ARGSUSED*/
 int copyfile(arg, tr)
-char *arg;
+CONST char *arg;
 int tr;
 {
 	int order;
@@ -1639,7 +1842,7 @@ int tr;
 	destpath = getdestdir(COPYD_K, arg);
 #else
 # ifdef	_NODOSDRIVE
-	destpath = (tr) ? tree(1, (int *)1) : getdestdir(COPYD_K, arg);
+	destpath = (tr) ? tree(1, NULL) : getdestdir(COPYD_K, arg);
 # else
 	destpath = (tr) ? tree(1, &destdrive) : getdestdir(COPYD_K, arg);
 # endif
@@ -1649,7 +1852,10 @@ int tr;
 	destdir = NULL;
 	if (mark > 0
 	|| (!isdir(&(filelist[filepos])) || islink(&(filelist[filepos])))) {
-		preparecopy();
+		if (preparecopy(0, mark) < 0) {
+			free(destpath);
+			return(FNC_CANCEL);
+		}
 		VOID_C applyfile(safecopy, ENDCP_K);
 	}
 #ifdef	_NOEXTRACOPY
@@ -1660,7 +1866,10 @@ int tr;
 	}
 #endif
 	else {
-		preparecopy();
+		if (preparecopy(1, 1) < 0) {
+			free(destpath);
+			return(FNC_CANCEL);
+		}
 		order = (islowerdir()) ? ORD_LOWER : ORD_NORMAL;
 		VOID_C applydir(filelist[filepos].name, safecopy,
 			cpdir, touchdir, order, ENDCP_K);
@@ -1673,10 +1882,11 @@ int tr;
 	forwardpath = NULL;
 #endif
 #ifndef	_NODOSDRIVE
-	if (destdrive >= 0) shutdrv(destdrive);
+	shutdrv(destdrive);
+	destdrive = -1;
 #endif
 #if	!defined (_NOEXTRACOPY) && !defined (_NODOSDRIVE)
-	if (forwarddrive >= 0) shutdrv(forwarddrive);
+	shutdrv(forwarddrive);
 	forwarddrive = -1;
 #endif
 
@@ -1685,14 +1895,11 @@ int tr;
 
 /*ARGSUSED*/
 int movefile(arg, tr)
-char *arg;
+CONST char *arg;
 int tr;
 {
 #ifdef	_USEDOSEMU
 	char path[MAXPATHLEN];
-#endif
-#ifndef	_NOEXTRACOPY
-	int order;
 #endif
 	int ret;
 
@@ -1704,7 +1911,7 @@ int tr;
 	destpath = getdestdir(MOVED_K, arg);
 #else
 # ifdef	_NODOSDRIVE
-	destpath = (tr) ? tree(1, (int *)1) : getdestdir(MOVED_K, arg);
+	destpath = (tr) ? tree(1, NULL) : getdestdir(MOVED_K, arg);
 # else
 	destpath = (tr) ? tree(1, &destdrive) : getdestdir(MOVED_K, arg);
 # endif
@@ -1717,7 +1924,10 @@ int tr;
 	}
 	destdir = NULL;
 	if (mark > 0) {
-		preparemove();
+		if (preparemove(0, mark) < 0) {
+			free(destpath);
+			return(FNC_CANCEL);
+		}
 		filepos = applyfile(safemove, ENDMV_K);
 	}
 	else if (islowerdir()) warning(EINVAL, filelist[filepos].name);
@@ -1729,18 +1939,23 @@ int tr;
 		else ret = isxdev(fnodospath(path, filepos), NULL);
 		if (ret < 0) warning(-1, filelist[filepos].name);
 		else if (ret) {
-			preparemove();
+			if (preparemove(1, 1) < 0) {
+				free(destpath);
+				return(FNC_CANCEL);
+			}
 # ifndef	_NODOSDRIVE
 			if (dospath3(nullstr)) waitmes();
 # endif
-			order = (islowerdir()) ? ORD_LOWER : ORD_NORMAL;
 			ret = applydir(filelist[filepos].name, safemove,
-				mvdir1, mvdir2, order, ENDMV_K);
+				mvdir1, mvdir2, ORD_NORMAL, ENDMV_K);
 		}
 		else
 #endif	/* !_NOEXTRACOPY */
 		{
-			preparemove();
+			if (preparemove(0, 1) < 0) {
+				free(destpath);
+				return(FNC_CANCEL);
+			}
 			ret = safemove(fnodospath(path, filepos));
 			if (ret == APL_ERROR)
 				warning(-1, filelist[filepos].name);
@@ -1756,10 +1971,11 @@ int tr;
 	forwardpath = NULL;
 #endif
 #ifndef	_NODOSDRIVE
-	if (destdrive >= 0) shutdrv(destdrive);
+	shutdrv(destdrive);
+	destdrive = -1;
 #endif
 #if	!defined (_NOEXTRACOPY) && !defined (_NODOSDRIVE)
-	if (forwarddrive >= 0) shutdrv(forwarddrive);
+	shutdrv(forwarddrive);
 	forwarddrive = -1;
 #endif
 
@@ -1767,7 +1983,7 @@ int tr;
 }
 
 static int forcecpfile(path)
-char *path;
+CONST char *path;
 {
 	struct stat st;
 	char dest[MAXPATHLEN];
@@ -1778,14 +1994,15 @@ char *path;
 	st.st_flags = (u_long)-1;
 #endif
 
-	st.st_nlink = (TCH_MODE | TCH_UID | TCH_GID | TCH_ATIME | TCH_MTIME);
+	st.st_nlink = (TCH_MODE | TCH_UID | TCH_GID
+		| TCH_ATIME | TCH_MTIME | TCH_IGNOREERR);
 	if (touchfile(dest, &st) < 0) return(APL_ERROR);
 
 	return(APL_OK);
 }
 
 static int forcecpdir(path)
-char *path;
+CONST char *path;
 {
 	struct stat st;
 	char dest[MAXPATHLEN];
@@ -1808,7 +2025,7 @@ char *path;
 }
 
 static int forcetouchdir(path)
-char *path;
+CONST char *path;
 {
 	struct stat st;
 	char dest[MAXPATHLEN];
@@ -1819,7 +2036,8 @@ char *path;
 	st.st_flags = (u_long)-1;
 #endif
 
-	st.st_nlink = (TCH_MODE | TCH_UID | TCH_GID | TCH_ATIME | TCH_MTIME);
+	st.st_nlink = (TCH_MODE | TCH_UID | TCH_GID
+		| TCH_ATIME | TCH_MTIME | TCH_IGNOREERR);
 	if (touchfile(dest, &st) < 0) return(APL_ERROR);
 
 	return(APL_OK);
