@@ -19,6 +19,9 @@ extern char *unixgetcurdir __P_((char *, int));
 #endif
 
 extern int mark;
+#if	FD >= 2
+extern int sorttype;
+#endif
 extern int stackdepth;
 #ifndef	_NOTRADLAYOUT
 extern int tradlayout;
@@ -27,8 +30,6 @@ extern int sizeinfo;
 extern off_t marksize;
 extern namelist filestack[];
 extern char fullpath[];
-extern char typesymlist[];
-extern u_short typelist[];
 #ifndef	_NOCOLOR
 extern int ansicolor;
 #endif
@@ -83,9 +84,9 @@ static char *form_tar[] = {
 #endif	/* !MSDOS */
 	NULL
 };
-#define	PM_LHA	form_lha, ign_lha, 0, 0
-#define	PM_TAR	form_tar, NULL, 0, 0
-#define	PM_NULL	NULL, NULL, 0, 0
+#define	PM_LHA	form_lha, ign_lha, NULL, 0, 0
+#define	PM_TAR	form_tar, NULL, NULL, 0, 0
+#define	PM_NULL	NULL, NULL, NULL, 0, 0
 #define	LINESEP	'\n'
 #define	MAXSCORE	256
 #define	isarchbr(l)	((l) -> format)
@@ -146,52 +147,61 @@ static char *form_tar[] = {
 #define	isarchbr(l)	(((l) -> topskip) < (u_char)255)
 #endif	/* FD < 2 */
 
-#define	iswhitespace(c)	((c) == ' ' || (c) == '\t')
+#define	IGNORETYPESYM	"-VMNLhC"
+#define	SYMLINKSTR	"->"
+#define	GNULINKSTR	"link to"
 
 #ifndef	_NOBROWSE
-static VOID NEAR copyargvar __P_((int, char **));
-static VOID NEAR pushbrowsevar __P_((char *));
+static VOID NEAR copyargvar __P_((int, char *CONST *));
+static VOID NEAR pushbrowsevar __P_((CONST char *));
 static VOID NEAR popbrowsevar __P_((VOID_A));
 #endif
 static VOID NEAR pusharchdupl __P_((VOID_A));
-#ifdef	_NOBROWSE
-#define	poparchdupl	escapearch
-#else
 static VOID NEAR poparchdupl __P_((VOID_A));
-#endif
-static int NEAR readattr __P_((namelist *, char *));
+static int NEAR readattr __P_((namelist *, CONST char *));
 #if	FD >= 2
-static char *NEAR checkspace __P_((char *, int *));
+static char *NEAR checkspace __P_((CONST char *, int *));
 # ifdef	_NOKANJIFCONV
 #define	readfname	strndup2
 # else
-static char *NEAR readfname __P_((char *, int));
+static char *NEAR readfname __P_((CONST char *, int));
 # endif
-static int NEAR readfileent __P_((namelist *, char *, char *, int));
+# ifndef	NOSYMLINK
+static char *NEAR readlinkname __P_((CONST char *, CONST char *));
+# endif
+static int NEAR readfileent __P_((namelist *,
+		CONST char *, CONST char *, int));
 #else	/* FD < 2 */
-static int NEAR countfield __P_((char *, u_char [], int, int *));
-static char *NEAR getfield __P_((char *, char *, int, launchtable *, int));
-static int NEAR readfileent __P_((namelist *, char *, launchtable *, int));
+static int NEAR countfield __P_((CONST char *, CONST u_char [], int, int *));
+static char *NEAR getfield __P_((char *, CONST char *,
+		int, launchtable *, int));
+static int NEAR readfileent __P_((namelist *,
+		CONST char *, launchtable *, int));
 #endif	/* FD < 2 */
-static int NEAR dircmp __P_((char *, char *));
-static int NEAR dirmatchlen __P_((char *, char *));
+static int NEAR dircmp __P_((CONST char *, CONST char *));
+static int NEAR dirmatchlen __P_((CONST char *, CONST char *));
 static char *NEAR pseudodir __P_((namelist *));
 static VOID NEAR Xwaitmes __P_((VOID_A));
 #if	FD >= 2
-static char **NEAR decodevar __P_((char **));
-static int NEAR matchlist __P_((char *, char **));
+static char **NEAR decodevar __P_((char *CONST *));
+static int NEAR matchlist __P_((CONST char *, char *CONST *));
 #endif
 static int NEAR parsearchive __P_((FILE *, launchtable *, namelist *, int *));
 static VOID NEAR unpackerror __P_((VOID_A));
-static int NEAR readarchive __P_((char *, launchtable *, int));
-static char *NEAR searcharcdir __P_((char *, int));
+static int NEAR readarchive __P_((CONST char *, launchtable *, int));
+static char *NEAR searcharcdir __P_((CONST char *, int));
 static char *NEAR archoutdir __P_((VOID_A));
-static int NEAR undertmp __P_((char *));
+static int NEAR undertmp __P_((CONST char *));
 #ifdef	_NODOSDRIVE
-static char *NEAR genfullpath __P_((char *, char *, char *));
+static char *NEAR genfullpath __P_((char *, CONST char *, CONST char *));
 #else
-static char *NEAR genfullpath __P_((char *, char *, char *, char *));
-static int NEAR archdostmpdir __P_((char *, char **, char *));
+static char *NEAR genfullpath __P_((char *,
+		CONST char *, CONST char *, CONST char *));
+static int NEAR archdostmpdir __P_((char *, char **, CONST char *));
+#endif
+#ifndef	NOSYMLINK
+static int NEAR archrealpath __P_((char *, char *));
+static int NEAR unpacklink __P_((namelist *, CONST char *));
 #endif
 
 int maxlaunch = 0;
@@ -211,7 +221,9 @@ launchtable launchlist[MAXLAUNCHTABLE] = {
 	{"*.tar.Z",	"zcat %C|tar tvf -",	PM_TAR, 0},
 	{"*.tar.gz",	"gzip -cd %C|tar tvf -",	PM_TAR, 0},
 	{"*.tar.bz2",	"bzip2 -cd %C|tar tvf -",	PM_TAR, 0},
+# ifndef	CYGWIN
 	{"*.taZ",	"zcat %C|tar tvf -",	PM_TAR, 0},
+# endif
 	{"*.taz",	"gzip -cd %C|tar tvf -",	PM_TAR, 0},
 	{"*.tgz",	"gzip -cd %C|tar tvf -",	PM_TAR, 0},
 #endif	/* !MSDOS */
@@ -240,8 +252,10 @@ archivetable archivelist[MAXARCHIVETABLE] = {
 					"gzip -cd %C|tar xf - %TA", 0},
 	{"*.tar.bz2",	"tar cf - %T|bzip2 -c > %C",
 					"bzip2 -cd %C|tar xf - %TA", 0},
+# ifndef	CYGWIN
 	{"*.taZ",	"tar cf - %T|compress -c > %C",
 					"zcat %C|tar xf - %TA", 0},
+# endif
 	{"*.taz",	"tar cf - %T|gzip -c > %C",
 					"gzip -cd %C|tar xf - %TA", 0},
 	{"*.tgz",	"tar cf - %T|gzip -c > %C",
@@ -254,8 +268,15 @@ char archivedir[MAXPATHLEN];
 char **browsevar = NULL;
 #endif
 
+#ifndef	NOSYMLINK
+static CONST strtable linklist[] = {
+	{strsize(SYMLINKSTR), SYMLINKSTR},
+	{strsize(GNULINKSTR), GNULINKSTR},
+};
+#define	LINKLISTSIZ	arraysize(linklist)
+#endif	/* !NOSYMLINK */
 #if	FD >= 2
-static char *autoformat[] = {
+static CONST char *autoformat[] = {
 # if	MSDOS
 	"%*f\n%s %x %x %y-%m-%d %t %a",		/* LHa (v) */
 	" %*f\n%s %x %x %d-%m-%y %t %a",	/* RAR (v) */
@@ -290,7 +311,7 @@ static char *autoformat[] = {
 #ifndef	_NOBROWSE
 static VOID NEAR copyargvar(argc, argv)
 int argc;
-char **argv;
+char *CONST *argv;
 {
 	int i;
 
@@ -301,7 +322,7 @@ char **argv;
 }
 
 static VOID NEAR pushbrowsevar(s)
-char *s;
+CONST char *s;
 {
 	int i;
 
@@ -361,15 +382,15 @@ static VOID NEAR pusharchdupl(VOID_A)
 }
 
 VOID escapearch(VOID_A)
-#ifndef	_NOBROWSE
 {
+#ifndef	_NOBROWSE
 	if (!archduplp) return;
 	if (browselist) popbrowsevar();
+#endif
 	poparchdupl();
 }
 
 static VOID NEAR poparchdupl(VOID_A)
-#endif	/* !_NOBROWSE */
 {
 #ifndef	_NODOSDRIVE
 	int drive;
@@ -454,23 +475,20 @@ static VOID NEAR poparchdupl(VOID_A)
 
 static int NEAR readattr(tmp, buf)
 namelist *tmp;
-char *buf;
+CONST char *buf;
 {
-	int i, c, len;
-	u_int mode;
+	int i, len;
+	u_int n, mode;
 
 	len = strlen(buf);
 	if (len < 9) {
 		mode = (tmp -> st_mode & S_IFMT) | S_IREAD_ALL | S_IWRITE_ALL;
 		while (--len >= 0) {
-			c = tolower2(buf[len]);
-			for (i = 0; typesymlist[i]; i++)
-				if (c == typesymlist[i]) break;
-			if (typesymlist[i]) {
+			if ((n = getfmode(buf[len])) != (u_int)-1) {
 				mode &= ~S_IFMT;
-				mode |= typelist[i];
+				mode |= n;
 			}
-			else switch (c) {
+			else switch (tolower2(buf[len])) {
 				case 'a':
 					mode |= S_ISVTX;
 					break;
@@ -537,13 +555,11 @@ char *buf;
 		else if (buf[i] == 'T') mode |= S_ISVTX;
 		else if (buf[i] != '-') return(0);
 
-		if (len >= 10 && buf[len - 10] != '-') {
-			c = tolower2(buf[len - 10]);
-			for (i = 0; typesymlist[i]; i++)
-				if (c == typesymlist[i]) break;
-			if (!typesymlist[i]) return(0);
+		if (len >= 10 && !strchr(IGNORETYPESYM, buf[len - 10])) {
+			n = getfmode(buf[len - 10]);
+			if (n == (u_int)-1) return(0);
 			mode &= ~S_IFMT;
-			mode |= typelist[i];
+			mode |= n;
 		}
 	}
 	tmp -> st_mode = mode;
@@ -553,19 +569,19 @@ char *buf;
 
 #if	FD >= 2
 static char *NEAR checkspace(s, scorep)
-char *s;
+CONST char *s;
 int *scorep;
 {
 	int i, len;
 
-	if (iswhitespace(*s)) {
+	if (isblank2(*s)) {
 		(*scorep)++;
-		for (s++; iswhitespace(*s); s++);
+		s = skipspace(&(s[1]));
 	}
 	for (i = 0; s[len = i]; i++) {
-		if (iswhitespace(s[i])) {
+		if (isblank2(s[i])) {
 			(*scorep)++;
-			for (i++; iswhitespace(s[i]); i++);
+			for (i++; isblank2(s[i]); i++);
 			if (!s[i]) break;
 			*scorep += 4;
 		}
@@ -576,7 +592,7 @@ int *scorep;
 
 # ifndef	_NOKANJIFCONV
 static char *NEAR readfname(s, len)
-char *s;
+CONST char *s;
 int len;
 {
 	char *cp, *tmp;
@@ -589,9 +605,27 @@ int len;
 }
 # endif	/* !_NOKANJIFCONV */
 
+# ifndef	NOSYMLINK
+static char *NEAR readlinkname(s, eol)
+CONST char *s, *eol;
+{
+	int i;
+
+	for (i = 0; i < LINKLISTSIZ; i++) {
+		if (&(s[linklist[i].no]) > eol) continue;
+		if (!strncmp(s, linklist[i].str, linklist[i].no)) {
+			s += linklist[i].no;
+			return(skipspace(s));
+		}
+	}
+
+	return(NULL);
+}
+# endif	/* !NOSYMLINK */
+
 static int NEAR readfileent(tmp, line, form, skip)
 namelist *tmp;
-char *line, *form;
+CONST char *line, *form;
 int skip;
 {
 # ifndef	NOUID
@@ -599,12 +633,16 @@ int skip;
 	gidtable *gp;
 	uid_t uid;
 	gid_t gid;
+# endif
+# ifndef	NOSYMLINK
+	CONST char *line2, *lname;
 # endif
 	time_t now;
 	struct tm tm, *tp;
 	off_t n;
 	int i, ch, l, len, hit, err, err2, score;
-	char *cp, *s, *buf, *rawbuf;
+	CONST char *s;
+	char *cp, *buf, *eol, *rawbuf;
 
 	if (skip && skip > strlen(form)) return(-1);
 
@@ -623,14 +661,14 @@ int skip;
 	tm.tm_hour = tm.tm_min = tm.tm_sec = 0;
 	score = 0;
 
-	while (*form) if (iswhitespace(*form)) {
-		while (iswhitespace(*line)) line++;
+	while (*form) if (isblank2(*form)) {
+		line = skipspace(line);
 		form++;
 	}
 	else if (*form == '\n') {
 		if (*line != *(form++)) score += 20;
 		else line++;
-		while (iswhitespace(*line)) line++;
+		line = skipspace(line);
 	}
 	else if (*form != '%' || *(++form) == '%') {
 		ch = *(form++);
@@ -679,14 +717,14 @@ int skip;
 		}
 
 		if (len < 0) {
-			if (!iswhitespace(*form)
+			if (!isblank2(*form)
 			&& (*form != '%' || *(form + 1) == '%'))
 				ch = *form;
 			for (len = 0; line[len]; len++) {
 				if (ch) {
 					if (ch == line[len]) break;
 				}
-				else if (iswhitespace(line[len])) break;
+				else if (isblank2(line[len])) break;
 			}
 		}
 
@@ -785,32 +823,37 @@ int skip;
 				tm.tm_sec = i;
 				break;
 			case 'f':
+				eol = &(rawbuf[len]);
+# ifndef	NOSYMLINK
+				lname = NULL;
+# endif
 				for (i = 0; i < len; i++) {
+					cp = &(rawbuf[i]);
 # if	MSDOS
 #  ifdef	BSPATHDELIM
-					if (rawbuf[i] == '/') rawbuf[i] = _SC_;
+					if (*cp == '/') *cp = _SC_;
 #  else
-					if (rawbuf[i] == '\\')
-						rawbuf[i] = _SC_;
+					if (*cp == '\\') *cp = _SC_;
 					if (iskanji1(rawbuf, i)) {
 						i++;
 						continue;
 					}
 #  endif
 # endif	/* MSDOS */
-					if (!iswhitespace(rawbuf[i])) continue;
-					cp = &(rawbuf[i]);
-					for (cp++; iswhitespace(*cp); cp++);
-					if (cp >= &(rawbuf[len])) {
+					if (!isblank2(*cp)) continue;
+					cp = skipspace(&(cp[1]));
+					if (cp >= eol) {
 						if (!ch) i = len;
 						break;
 					}
 # ifndef	NOSYMLINK
-					if (cp[0] == '-' && cp[1] == '>')
+					if ((lname = readlinkname(cp, eol)))
 						break;
 # endif
 				}
-				cp = &(line[i]);
+# ifndef	NOSYMLINK
+				line2 = &(line[i]);
+# endif
 				if (isdelim(rawbuf, i - 1)) {
 					tmp -> st_mode &= ~S_IFMT;
 					tmp -> st_mode |= S_IFDIR;
@@ -823,17 +866,17 @@ int skip;
 				hit++;
 				err = 0;
 # ifndef	NOSYMLINK
-				while (iswhitespace(*cp)) cp++;
-				if (ch && cp >= &(line[len])) break;
-				if (cp[0] != '-' || cp[1] != '>') break;
-				for (cp += 2; iswhitespace(*cp); cp++);
-				for (i = 0; cp[i]; i++)
-					if (iswhitespace(cp[i])) break;
+				if (!lname) break;
+				line2 = skipspace(line2);
+				if (ch && line2 >= &(line[len])) break;
+				lname = &(line[lname - rawbuf]);
+				for (i = 0; lname[i]; i++)
+					if (isblank2(lname[i])) break;
 				if (i) {
 					if (tmp -> linkname)
 						free(tmp -> linkname);
-					tmp -> linkname = readfname(cp, i);
-					i = &(cp[i]) - line;
+					tmp -> linkname = readfname(lname, i);
+					i = &(lname[i]) - line;
 					if (i > len) len = i;
 				}
 # endif
@@ -862,7 +905,7 @@ int skip;
 		else if (hit <= err2) score += err2;
 		score += err;
 		line += len;
-		if (!ch) while (iswhitespace(*line)) line++;
+		if (!ch) line = skipspace(line);
 	}
 
 	if (score >= MAXSCORE || !(tmp -> name) || !*(tmp -> name)) {
@@ -917,8 +960,8 @@ int skip;
 #else	/* FD < 2 */
 
 static int NEAR countfield(line, sep, field, eolp)
-char *line;
-u_char sep[];
+CONST char *line;
+CONST u_char sep[];
 int field, *eolp;
 {
 	int i, j, f, s, sp;
@@ -932,12 +975,12 @@ int field, *eolp;
 			j++;
 			sp = (j < MAXLAUNCHSEP) ? (int)sep[j] : 255;
 			for (; sp == 255 || i < sp; i++)
-				if (!iswhitespace(line[i])) break;
+				if (!isblank2(line[i])) break;
 			if (f < 0) f = 1;
 			else f++;
 			s = 0;
 		}
-		else if (iswhitespace(line[i])) s = 1;
+		else if (isblank2(line[i])) s = 1;
 		else if (s) {
 			f++;
 			s = 0;
@@ -949,7 +992,7 @@ int field, *eolp;
 			if (eolp) {
 				for (j = i; line[j]; j++) {
 					if ((sp < 255 && j >= sp)
-					|| iswhitespace(line[j]))
+					|| isblank2(line[j]))
 						break;
 				}
 				*eolp = j;
@@ -962,12 +1005,14 @@ int field, *eolp;
 }
 
 static char *NEAR getfield(buf, line, skip, list, no)
-char *buf, *line;
+char *buf;
+CONST char *line;
 int skip;
 launchtable *list;
 int no;
 {
-	char *cp, *tmp;
+	CONST char *cp;
+	char *tmp;
 	int i, f, eol;
 
 	*buf = '\0';
@@ -1000,7 +1045,7 @@ int no;
 
 static int NEAR readfileent(tmp, line, list, max)
 namelist *tmp;
-char *line;
+CONST char *line;
 launchtable *list;
 int max;
 {
@@ -1137,7 +1182,7 @@ int max;
 #endif	/* FD < 2 */
 
 VOID archbar(file, dir)
-char *file, *dir;
+CONST char *file, *dir;
 {
 	char *arch;
 	int len;
@@ -1227,7 +1272,7 @@ char *file, *dir;
 }
 
 static int NEAR dircmp(s1, s2)
-char *s1, *s2;
+CONST char *s1, *s2;
 {
 	int i, j;
 
@@ -1248,7 +1293,7 @@ char *s1, *s2;
 }
 
 static int NEAR dirmatchlen(s1, s2)
-char *s1, *s2;
+CONST char *s1, *s2;
 {
 	int i, j;
 
@@ -1332,7 +1377,7 @@ static VOID NEAR Xwaitmes(VOID_A)
 
 #if	FD >= 2
 static char **NEAR decodevar(argv)
-char **argv;
+char *CONST *argv;
 {
 	char **new;
 	int n, max;
@@ -1346,13 +1391,15 @@ char **argv;
 }
 
 static int NEAR matchlist(s, argv)
-char *s, **argv;
+CONST char *s;
+char *CONST *argv;
 {
 # ifndef	PATHNOCASE
 	int duppathignorecase;
 # endif
 	reg_t *re;
-	char *s1, *s2;
+	CONST char *s1, *s2;
+	char *new;
 	int i, len, ret;
 
 	if (!argv) return(0);
@@ -1365,18 +1412,18 @@ char *s, **argv;
 	for (i = 0; !ret && argv[i]; i++) {
 		s1 = argv[i];
 		s2 = s;
-		if (!iswhitespace(s1[0])) while (iswhitespace(*s2)) s2++;
+		if (!isblank2(s1[0])) s2 = skipspace(s2);
 		len = strlen(s1);
-		if (!len || iswhitespace(s1[len - 1])) s2 = strdup2(s2);
+		if (!len || isblank2(s1[len - 1])) new = strdup2(s2);
 		else {
 			len = strlen(s2);
 			for (len--; len >= 0; len--)
-				if (!iswhitespace(s2[len])) break;
-			s2 = strndup2(s2, ++len);
+				if (!isblank2(s2[len])) break;
+			new = strndup2(s2, ++len);
 		}
 		re = regexp_init(s1, -1);
-		ret = regexp_exec(re, s2, 0);
-		free(s2);
+		ret = regexp_exec(re, new, 0);
+		free(new);
 		regexp_free(re);
 	}
 # ifndef	PATHNOCASE
@@ -1398,7 +1445,8 @@ int *linenop;
 	static char **lign = NULL;
 	static char **lerr = NULL;
 	namelist tmp;
-	char *form, *form0;
+	CONST char *form;
+	char *form0, *form2;
 	short *scorelist;
 	int nf, na, ret, skip;
 #else
@@ -1434,7 +1482,7 @@ int *linenop;
 		lvar[++nline] = NULL;
 	}
 
-	if (isttyiomode && intrkey()) return(-2);
+	if (isttyiomode && intrkey(K_ESC)) return(-2);
 #if	FD >= 2
 	if (matchlist(lvar[0], lign)) {
 		free(lvar[0]);
@@ -1467,7 +1515,10 @@ int *linenop;
 	needline = 0;
 	for (;;) {
 #if	FD >= 2
-		if (formlist[nf]) form = formlist[nf];
+# ifdef	FAKEUNINIT
+		form2 = NULL;		/* fake for -Wuninitialized */
+# endif
+		if (formlist[nf]) form = form2 = formlist[nf];
 		else if (scorelist[0] < MAXSCORE) {
 			ret = scorelist[0] + 1;
 			score = 0;
@@ -1555,7 +1606,7 @@ int *linenop;
 				break;
 			}
 			scorelist[i] = score;
-			formlist[i] = form;
+			formlist[i] = form2;
 			nf++;
 		}
 
@@ -1626,7 +1677,7 @@ static VOID NEAR unpackerror(VOID_A)
 }
 
 static int NEAR readarchive(file, list, flags)
-char *file;
+CONST char *file;
 launchtable *list;
 int flags;
 {
@@ -1724,7 +1775,7 @@ int flags;
 	if (maxfile <= 1) {
 		maxfile = 0;
 		free(filelist[0].name);
-		filelist[0].name = NOFIL_K;
+		filelist[0].name = (char *)NOFIL_K;
 #ifndef	NOSYMLINK
 		if (filelist[0].linkname) {
 			free(filelist[0].linkname);
@@ -1751,8 +1802,8 @@ int flags;
 }
 
 VOID copyarcf(re, arcre)
-reg_t *re;
-char *arcre;
+CONST reg_t *re;
+CONST char *arcre;
 {
 	char *cp, *tmp;
 	int i, j, n, len, parent;
@@ -1813,7 +1864,7 @@ char *arcre;
 }
 
 static char *NEAR searcharcdir(file, flen)
-char *file;
+CONST char *file;
 int flen;
 {
 	char *cp, *tmp;
@@ -1876,10 +1927,11 @@ static char *NEAR archoutdir(VOID_A)
 	return(file);
 }
 
-char *archchdir(path)
-char *path;
+CONST char *archchdir(path)
+CONST char *path;
 {
-	char *cp, *file, duparcdir[MAXPATHLEN];
+	CONST char *cp, *file;
+	char *tmp, duparcdir[MAXPATHLEN];
 	int len;
 
 	if (findpattern) free(findpattern);
@@ -1945,8 +1997,8 @@ char *path;
 		cp = path;
 		if (len == 2 && path[0] == '.' && path[1] == '.') cp = nullstr;
 		if (searcharcdir(cp, len)) {
-			if (*(cp = archivedir)) cp = strcatdelim(archivedir);
-			strncpy2(cp, path, len);
+			if (*(tmp = archivedir)) tmp = strcatdelim(archivedir);
+			strncpy2(tmp, path, len);
 			file = parentpath;
 		}
 		else if (*cp || !(file = archoutdir())) {
@@ -1965,11 +2017,12 @@ char *path;
 
 #ifndef	_NOCOMPLETE
 int completearch(path, flen, argc, argvp)
-char *path;
+CONST char *path;
 int flen, argc;
 char ***argvp;
 {
-	char *cp, *tmp, *new, *file, dir[MAXPATHLEN], duparcdir[MAXPATHLEN];
+	CONST char *cp, *file;
+	char *tmp, *new, dir[MAXPATHLEN], duparcdir[MAXPATHLEN];
 	int i, len, parent;
 
 # ifdef	_USEDOSPATH
@@ -2037,6 +2090,7 @@ int flags;
 #ifndef	_NODOSDRIVE
 	int drive;
 #endif
+	CONST char *cp;
 	char *tmpdir;
 	int i;
 
@@ -2085,10 +2139,9 @@ int flags;
 			archduplp -> v_archivedir = strdup2(archivedir);
 		}
 	}
-	archivefile = (filelist && filepos < maxfile)
-		? filelist[filepos].name : NULL;
-	if (!archivefile) archivefile = nullstr;
-	archivefile = strdup2(archivefile);
+	cp = (filelist && filepos < maxfile) ? filelist[filepos].name : NULL;
+	if (!cp) cp = nullstr;
+	archivefile = strdup2(cp);
 	*archivedir = '\0';
 	archtmpdir = tmpdir;
 #ifndef	_NODOSDRIVE
@@ -2101,7 +2154,11 @@ int flags;
 	launchp = list;
 	filelist = NULL;
 	maxent = 0;
+#if	FD >= 2
+	if (sorttype < 200) sorton = 0;
+#else
 	sorton = 0;
+#endif
 
 	if (readarchive(archivefile, launchp, flags) < 0) {
 		arcflist = NULL;
@@ -2170,7 +2227,7 @@ int launcher(VOID_A)
 }
 
 static int NEAR undertmp(path)
-char *path;
+CONST char *path;
 {
 	winvartable *wvp;
 	char *full, rpath[MAXPATHLEN], dir[MAXPATHLEN];
@@ -2189,25 +2246,27 @@ char *path;
 
 #ifdef	_NODOSDRIVE
 static char *NEAR genfullpath(path, file, full)
-char *path, *file, *full;
+char *path;
+CONST char *file, *full;
 #else
 static char *NEAR genfullpath(path, file, full, tmpdir)
-char *path, *file, *full, *tmpdir;
+char *path;
+CONST char *file, *full, *tmpdir;
 #endif
 {
 #if	MSDOS
+	char *tmp;
 	int drive;
 #endif
-	char *cp;
+	CONST char *cp;
 
 	cp = file;
-#ifdef	_USEDOSPATH
-# if	MSDOS
+#if	MSDOS
 	if ((drive = _dospath(cp))) cp += 2;
-# else
+#endif
+#ifdef	_USEDOSEMU
 	if (_dospath(cp)) cp += 2;
-# endif
-#endif	/* _USEDOSPATH */
+#endif
 
 #ifndef	_NODOSDRIVE
 	if (tmpdir) strcpy(path, tmpdir);
@@ -2223,10 +2282,10 @@ char *path, *file, *full, *tmpdir;
 	}
 #if	MSDOS
 	else if (drive) {
-		file = gendospath(path, drive, _SC_);
+		tmp = gendospath(path, drive, _SC_);
 		drive = toupper2(drive);
-		if (drive == toupper2(*full)) strcpy(file, &(full[3]));
-		else if (!unixgetcurdir(file, drive - 'A' + 1)) return(NULL);
+		if (drive == toupper2(*full)) strcpy(tmp, &(full[3]));
+		else if (!unixgetcurdir(tmp, drive - 'A' + 1)) return(NULL);
 	}
 #endif
 	else strcpy(path, full);
@@ -2237,7 +2296,8 @@ char *path, *file, *full, *tmpdir;
 
 #ifndef	_NODOSDRIVE
 static int NEAR archdostmpdir(path, dirp, full)
-char *path, **dirp, *full;
+char *path, **dirp;
+CONST char *full;
 {
 	char *cp, dupfullpath[MAXPATHLEN];
 	int drive;
@@ -2270,7 +2330,7 @@ char *path, **dirp, *full;
 #endif	/* !_NODOSDRIVE */
 
 int pack(arc)
-char *arc;
+CONST char *arc;
 {
 #ifndef	_NODOSDRIVE
 	char *dest, *tmpdest;
@@ -2370,18 +2430,17 @@ char *arc;
 
 /*ARGSUSED*/
 int unpack(arc, dir, arg, tr, flags)
-char *arc, *dir;
-char *arg;
+CONST char *arc, *dir, *arg;
 int tr, flags;
 {
 #ifndef	_NODOSDRIVE
 	winvartable *wvp;
 	namelist alist[1], *dupfilelist;
 	char *full, *dest, *tmpdest;
-	int dd, drive, dupmaxfile, dupfilepos;
+	int drive, dupmaxfile, dupfilepos;
 #endif
 	reg_t *re;
-	char *cp, *tmpdir, path[MAXPATHLEN];
+	char *cp, *new, *tmpdir, path[MAXPATHLEN];
 	int i, n, ret;
 
 #ifndef	_NOBROWSE
@@ -2408,20 +2467,13 @@ int tr, flags;
 	if (dir) strcpy(path, dir);
 	else for (;;) {
 #ifndef	_NOTREE
-		if (tr) {
-# ifdef	_NODOSDRIVE
-			dir = tree(0, (int *)1);
-# else
-			dir = tree(0, &dd);
-			if (dd >= 0) shutdrv(dd);
-# endif
-		}
+		if (tr) dir = new = tree(0, NULL);
 		else
 #endif	/* !_NOTREE */
 		{
-			if (arg && *arg) dir = strdup2(arg);
-			else dir = inputstr(UNPAC_K, 1, -1, NULL, HST_PATH);
-			dir = evalpath(dir, 0);
+			if (arg && *arg) new = strdup2(arg);
+			else new = inputstr(UNPAC_K, 1, -1, NULL, HST_PATH);
+			dir = new = evalpath(new, 0);
 		}
 		if (!dir) return(0);
 		if (!*dir) copycurpath(path);
@@ -2431,7 +2483,7 @@ int tr, flags;
 			if (_dospath(dir) && !dir[2]) copycurpath(cp);
 #endif
 		}
-		free(dir);
+		free(new);
 		dir = NULL;
 
 		if (!undertmp(path)) break;
@@ -2466,7 +2518,7 @@ int tr, flags;
 	dupfilelist = filelist;
 	dupmaxfile = maxfile;
 	dupfilepos = filepos;
-	alist[0].name = arc;
+	alist[0].name = (char *)arc;
 # ifndef	NOSYMLINK
 	alist[0].linkname = NULL;
 # endif
@@ -2515,6 +2567,77 @@ int tr, flags;
 	return((ret) ? 0 : 1);
 }
 
+#ifndef	NOSYMLINK
+static int NEAR archrealpath(path, resolved)
+char *path, *resolved;
+{
+	char *cp, tmp[MAXPATHLEN];
+	int n;
+
+	if (!*path || (n = isdotdir(path)) == 2) /*EMPTY*/;
+	else if (n == 1) {
+		cp = strrdelim(resolved, 0);
+		if (!cp) return(-1);
+		*cp = '\0';
+	}
+	else if ((cp = strdelim(path, 0))) {
+		strncpy2(tmp, path, cp - path);
+		if (archrealpath(tmp, resolved) < 0) return(-1);
+		if (archrealpath(++cp, resolved) < 0) return(-1);
+	}
+	else {
+		cp = strcatdelim(resolved);
+		strncpy2(cp, path, MAXPATHLEN - 1 - (cp - resolved));
+	}
+
+	return(0);
+}
+
+static int NEAR unpacklink(list, dir)
+namelist *list;
+CONST char *dir;
+{
+	namelist duplist;
+	char *cp, path[MAXPATHLEN], duparcdir[MAXPATHLEN];
+	int i, ret;
+
+	if (!(list -> linkname)) return(0);
+
+	if (!islink(list)) cp = list -> linkname;
+	else {
+		strcatdelim2(duparcdir, archivedir, list -> linkname);
+		cp = duparcdir;
+	}
+	*path = '\0';
+	if (archrealpath(cp, path) < 0) return(-1);
+
+	for (i = 0; i < maxarcf; i++) {
+		*duparcdir = '\0';
+		if (archrealpath(arcflist[i].name, duparcdir) < 0) continue;
+		if (!strcmp(path, duparcdir)) break;
+	}
+	if (i >= maxarcf) return(-1);
+
+	if (unpacklink(&(arcflist[i]), dir) < 0) return(-1);
+
+	strcpy(duparcdir, archivedir);
+	memcpy(&duplist, &(filelist[filepos]), sizeof(namelist));
+	memcpy(&(filelist[filepos]), &(arcflist[i]), sizeof(namelist));
+	if (!(cp = strrdelim(filelist[filepos].name, 0))) *archivedir = '\0';
+	else {
+		strncpy2(archivedir,
+			filelist[filepos].name, cp - filelist[filepos].name);
+		filelist[filepos].name = ++cp;
+	}
+	ret = unpack(archivefile, dir, NULL,
+		0, F_ARGSET | F_ISARCH | F_NOADDOPT);
+	memcpy(&(filelist[filepos]), &duplist, sizeof(namelist));
+	strcpy(archivedir, duparcdir);
+
+	return(ret);
+}
+#endif	/* !NOSYMLINK */
+
 char *tmpunpack(single)
 int single;
 {
@@ -2536,7 +2659,12 @@ int single;
 	for (i = 0; i < maxfile; i++)
 		if (ismark(&(filelist[i]))) filelist[i].tmpflags |= F_WSMRK;
 	dupmark = mark;
-	if (single) mark = 0;
+	if (single) {
+		mark = 0;
+#ifndef	NOSYMLINK
+		VOID_C unpacklink(&(filelist[filepos]), path);
+#endif
+	}
 	ret = unpack(archivefile, path, NULL,
 		0, F_ARGSET | F_ISARCH | F_NOADDOPT);
 	mark = dupmark;
@@ -2581,7 +2709,7 @@ int single;
 }
 
 int backup(dev)
-char *dev;
+CONST char *dev;
 {
 	macrostat st;
 	char *tmp;
@@ -2628,7 +2756,7 @@ char *dev;
 }
 
 int searcharc(regstr, flist, maxf, n)
-char *regstr;
+CONST char *regstr;
 namelist *flist;
 int maxf, n;
 {
