@@ -32,6 +32,11 @@
 
 #if	MSDOS && !defined (FD)
 #define	VOL_FAT32	"FAT32"
+# ifdef	DOUBLESLASH
+# define	MNTDIRSIZ	MAXPATHLEN
+# else
+# define	MNTDIRSIZ	(3 + 1)
+# endif
 #endif	/* MSDOS && !FD */
 
 #if	MSDOS
@@ -170,17 +175,17 @@ static int cc_intr = -1;
 #endif	/* !FD */
 
 #ifdef	FD
-extern DIR *Xopendir __P_((char *));
+extern DIR *Xopendir __P_((CONST char *));
 extern int Xclosedir __P_((DIR *));
 extern struct dirent *Xreaddir __P_((DIR *));
 extern char *Xgetwd __P_((char *));
-extern int Xstat __P_((char *, struct stat *));
-extern int Xlstat __P_((char *, struct stat *));
-extern int Xaccess __P_((char *, int));
-extern int Xchmod __P_((char *, int));
-extern int Xunlink __P_((char *));
-extern int Xrename __P_((char *, char *));
-extern int Xopen __P_((char *, int, int));
+extern int Xstat __P_((CONST char *, struct stat *));
+extern int Xlstat __P_((CONST char *, struct stat *));
+extern int Xaccess __P_((CONST char *, int));
+extern int Xchmod __P_((CONST char *, int));
+extern int Xunlink __P_((CONST char *));
+extern int Xrename __P_((CONST char *, CONST char *));
+extern int Xopen __P_((CONST char *, int, int));
 # ifdef	_NODOSDRIVE
 # define	Xclose(f)	((close(f)) ? -1 : 0)
 # define	Xlseek		lseek
@@ -188,13 +193,12 @@ extern int Xopen __P_((char *, int, int));
 extern int Xclose __P_((int));
 extern off_t Xlseek __P_((int, off_t, int));
 # endif	/* !_NODOSDRIVE */
-extern int Xmkdir __P_((char *, int));
-extern int Xrmdir __P_((char *));
-extern int stat2 __P_((char *, struct stat *));
-extern int chdir2 __P_((char *));
+extern int Xmkdir __P_((CONST char *, int));
+extern int Xrmdir __P_((CONST char *));
+extern int stat2 __P_((CONST char *, struct stat *));
 #else	/* !FD */
 # if	MSDOS
-extern DIR *Xopendir __P_((char *));
+extern DIR *Xopendir __P_((CONST char *));
 extern int Xclosedir __P_((DIR *));
 extern struct dirent *Xreaddir __P_((DIR *));
 # else
@@ -203,7 +207,7 @@ extern struct dirent *Xreaddir __P_((DIR *));
 # define	Xreaddir	readdir
 # endif
 # ifdef	DJGPP
-extern dos_putpath __P_((char *, int));
+extern int dos_putpath __P_((CONST char *, int));
 extern char *Xgetwd __P_((char *));
 # else	/* !DJGPP */
 #  ifdef	USEGETWD
@@ -213,7 +217,7 @@ extern char *Xgetwd __P_((char *));
 #  endif
 # endif	/* !DJGPP */
 # if	MSDOS
-extern int Xstat __P_((char *, struct stat *));
+extern int Xstat __P_((CONST char *, struct stat *));
 # define	Xlstat		Xstat
 # else
 # define	Xstat(p, s)	((stat(p, s)) ? -1 : 0)
@@ -234,7 +238,7 @@ extern int Xstat __P_((char *, struct stat *));
 #  ifdef	DJGPP
 #  define	Xmkdir(p, m)	((mkdir(p, m)) ? -1 : 0)
 #  else
-extern int Xmkdir __P_((char *, int));
+extern int Xmkdir __P_((CONST char *, int));
 #  endif
 # else
 # define	Xmkdir(p, m)	((mkdir(p, m)) ? -1 : 0)
@@ -242,14 +246,11 @@ extern int Xmkdir __P_((char *, int));
 #define	Xrmdir(p)	((rmdir(p)) ? -1 : 0)
 # if	MSDOS
 extern int intcall __P_((int, __dpmi_regs *, struct SREGS *));
-extern int chdir2 __P_((char *));
-# else
-# define	chdir2(p)	((chdir(p)) ? -1 : 0)
 # endif
 # ifdef	NOSYMLINK
 # define	stat2		Xstat
 # else
-extern int stat2 __P_((char *, struct stat *));
+extern int stat2 __P_((CONST char *, struct stat *));
 # endif
 #endif	/* !FD */
 
@@ -306,32 +307,46 @@ struct filestat_t {
 extern char *malloc2 __P_((ALLOC_T));
 extern char *realloc2 __P_((VOID_P, ALLOC_T));
 extern char *c_realloc __P_((char *, ALLOC_T, ALLOC_T *));
-extern char *strdup2 __P_((char *));
+extern char *strdup2 __P_((CONST char *));
+extern char *strncpy2 __P_((char *, CONST char *, int));
 
 #ifdef	FD
-extern int getinfofs __P_((char *, off_t *, off_t *, off_t *));
-extern char *realpath2 __P_((char *, char *, int));
-extern int touchfile __P_((char *, struct stat *));
+# ifdef	_USEDOSPATH
+extern int _dospath __P_((CONST char *));
+# endif
+# if	MSDOS
+extern int dospath3 __P_((CONST char *));
+# else
+extern int dospath __P_((CONST char *, char *));
+#define	dospath3(path)	dospath(path, NULL)
+# endif
+# ifndef	_NODOSDRIVE
+extern int preparedrv __P_((int, int *));
+extern VOID shutdrv __P_((int));
+# endif
+extern int getinfofs __P_((CONST char *, off_t *, off_t *, off_t *));
+extern char *realpath2 __P_((CONST char *, char *, int));
+extern int touchfile __P_((CONST char *, struct stat *));
 #ifndef	NODIRLOOP
-extern int issamebody __P_((char *, char *));
+extern int issamebody __P_((CONST char *, CONST char *));
 #endif
 #ifndef	NOSYMLINK
-extern int cpsymlink __P_((char *, char *));
+extern int cpsymlink __P_((CONST char *, CONST char *));
 #endif
-extern char *inputstr __P_((char *, int, int, char *, int));
+extern char *inputstr __P_((CONST char *, int, int, CONST char *, int));
 #else	/* !FD */
-static int NEAR getinfofs __P_((char *, off_t *, off_t *, off_t *));
+static int NEAR getinfofs __P_((CONST char *, off_t *, off_t *, off_t *));
 # if	MSDOS
-static char *NEAR realpath2 __P_((char *, char *, int));
+static char *NEAR realpath2 __P_((CONST char *, char *, int));
 static int NEAR putdostime __P_((u_short *, u_short *, time_t));
 #  ifdef	USEUTIME
-static int NEAR Xutime __P_((char *, struct utimbuf *));
+static int NEAR Xutime __P_((CONST char *, CONST struct utimbuf *));
 #  else
-static int NEAR Xutimes __P_((char *, struct timeval []));
+static int NEAR Xutimes __P_((CONST char *, CONST struct timeval *));
 #  endif
 # define	ttyiomode(n)
 # define	stdiomode()
-# define	getkey2(n)	getch();
+# define	getkey3(n, c)	getch();
 # else	/* !MSDOS */
 # define	realpath2(p, r, f) \
 				realpath(p, r)
@@ -340,26 +355,26 @@ static int NEAR Xutimes __P_((char *, struct timeval []));
 static VOID NEAR ttymode __P_((int));
 # define	ttyiomode(n)	(ttymode(1))
 # define	stdiomode()	(ttymode(0))
-static int NEAR getkey2 __P_((int));
+static int NEAR getkey3 __P_((int, int));
 # endif	/* !MSDOS */
-static int NEAR touchfile __P_((char *, struct stat *));
+static int NEAR touchfile __P_((CONST char *, struct stat *));
 #ifndef	NODIRLOOP
-static int NEAR issamebody __P_((char *, char *));
+static int NEAR issamebody __P_((CONST char *, CONST char *));
 #endif
 #ifndef	NOSYMLINK
-static int NEAR cpsymlink __P_((char *, char *));
+static int NEAR cpsymlink __P_((CONST char *, CONST char *));
 #endif
-static char *NEAR inputstr __P_((char *, int, int, char *, int));
+static char *NEAR inputstr __P_((CONST char *, int, int, CONST char *, int));
 #endif	/* !FD */
 
-static VOID NEAR doserror __P_((char *, int));
-static VOID NEAR dosperror __P_((char *));
+static VOID NEAR doserror __P_((CONST char *, int));
+static VOID NEAR dosperror __P_((CONST char *));
 static VOID NEAR fputsize __P_((off_t *, off_t *));
 static int NEAR inputkey __P_((VOID_A));
 static int cmpdirent __P_((CONST VOID_P, CONST VOID_P));
-static VOID NEAR evalenvopt __P_((char *, char *,
-		int (NEAR *)__P_((int, char *[]))));
-static int NEAR getdiropt __P_((int, char *[]));
+static VOID NEAR evalenvopt __P_((CONST char *, CONST char *,
+		int (NEAR *)__P_((int, char *CONST []))));
+static int NEAR getdiropt __P_((int, char *CONST []));
 static int NEAR showstr __P_((char *, int, int));
 #ifdef	MINIMUMSHELL
 static VOID NEAR showfname __P_((struct filestat_t *, off_t *));
@@ -385,14 +400,24 @@ static VOID NEAR dosdirfooter __P_((off_t *, int, int,
 static int NEAR dosdir __P_((reg_t *, off_t *, int *, int *,
 		off_t *, off_t *));
 #endif
-static int NEAR checkarg __P_((int, char *[]));
-static char *NEAR convwild __P_((char *, char *, char *, char *));
-static int NEAR getcopyopt __P_((int, char *[]));
-static int NEAR getbinmode __P_((char *, int));
-static int NEAR writeopen __P_((char *, char *));
+int doscomdir __P_((int, char *CONST []));
+static int NEAR checkarg __P_((int, char *CONST []));
+int doscommkdir __P_((int, char *CONST []));
+int doscomrmdir __P_((int, char *CONST []));
+int doscomerase __P_((int, char *CONST []));
+static char *NEAR convwild __P_((char *, CONST char *,
+		CONST char *, CONST char *));
+int doscomrename __P_((int, char *CONST []));
+static int NEAR getcopyopt __P_((int, char *CONST []));
+static int NEAR getbinmode __P_((CONST char *, int));
+static int NEAR writeopen __P_((CONST char *, CONST char *));
 static int NEAR textread __P_((int, u_char *, int, int));
 static int NEAR textclose __P_((int, int));
-static int NEAR doscopy __P_((char *, char *, struct stat *, int, int, int));
+static int NEAR doscopy __P_((CONST char *, CONST char *,
+		struct stat *, int, int, int));
+int doscomcopy __P_((int, char *CONST []));
+int doscomcls __P_((int, char *CONST []));
+int doscomtype __P_((int, char *CONST []));
 
 #define	BUFUNIT		32
 #define	DIRSORTFLAG	"NSEDGA"
@@ -406,7 +431,7 @@ static int NEAR doscopy __P_((char *, char *, struct stat *, int, int, int));
 #define	strerror2		strerror
 #else
 # ifndef	DECLERRLIST
-extern char *sys_errlist[];
+extern CONST char *CONST sys_errlist[];
 # endif
 #define	strerror2(n)		(char *)sys_errlist[n]
 #endif
@@ -451,17 +476,37 @@ static int copyflag = 0;
 
 #ifndef	FD
 static int NEAR getinfofs(path, totalp, freep, bsizep)
-char *path;
+CONST char *path;
 off_t *totalp, *freep, *bsizep;
 {
 # if	MSDOS
+#  ifdef	DOUBLESLASH
+	char *cp;
+	int len;
+#  endif
 	struct SREGS sreg;
 	__dpmi_regs reg;
-	char drv[4], buf[128];
-# endif
+	char drv[MNTDIRSIZ], buf[128];
+# else	/* !MSDOS */
+	struct stat st;
+# endif	/* !MSDOS */
 	statfs_t fsbuf;
+	int n;
 
+	n = 0;
 # if	MSDOS
+#  ifdef	DOUBLESLASH
+	if ((len = isdslash(path))) {
+		if ((cp = strdelim(&(path[len]), 0))) {
+			len = cp - path;
+			if (!(cp = strdelim(&(path[len + 1]), 0)))
+				cp += strlen(cp);
+			len = cp - path;
+		}
+		strncpy2(drv, path, len);
+	}
+	else
+#  endif
 	VOID_C gendospath(drv, toupper2(path[0]), _SC_);
 
 	reg.x.ax = 0x71a0;
@@ -488,15 +533,16 @@ off_t *totalp, *freep, *bsizep;
 		reg.x.di = PTR_OFF(&fsbuf, 0);
 		sreg.ds = PTR_SEG(path);
 		reg.x.dx = PTR_OFF(path, sizeof(fsbuf));
-		if (intcall(0x21, &reg, &sreg) < 0) return(-1);
-
+		if (intcall(0x21, &reg, &sreg) < 0) n = -1;
+		else {
 #  ifdef	DJGPP
-		dosmemget(__tb, sizeof(fsbuf), &fsbuf);
+			dosmemget(__tb, sizeof(fsbuf), &fsbuf);
 #  endif
-		*totalp = (off_t)(fsbuf.f_blocks);
-		*freep = (off_t)(fsbuf.f_bavail);
-		*bsizep = (off_t)(fsbuf.f_clustsize)
-			* (off_t)(fsbuf.f_sectsize);
+			*totalp = (off_t)(fsbuf.f_blocks);
+			*freep = (off_t)(fsbuf.f_bavail);
+			*bsizep = (off_t)(fsbuf.f_clustsize)
+				* (off_t)(fsbuf.f_sectsize);
+		}
 	}
 	else {
 		reg.x.ax = 0x3600;
@@ -504,31 +550,44 @@ off_t *totalp, *freep, *bsizep;
 		intcall(0x21, &reg, &sreg);
 		if (reg.x.ax == 0xffff) {
 			errno = ENOENT;
-			return(-1);
+			n = -1;
+		}
+		else {
+			*totalp = (off_t)(reg.x.dx);
+			*freep = (off_t)(reg.x.bx);
+			*bsizep = (off_t)(reg.x.ax) * (off_t)(reg.x.cx);
 		}
-
-		*totalp = (off_t)(reg.x.dx);
-		*freep = (off_t)(reg.x.bx);
-		*bsizep = (off_t)(reg.x.ax) * (off_t)(reg.x.cx);
 	}
 	if (!*totalp || !*bsizep || *totalp < *freep) {
 		errno = EIO;
-		return(-1);
+		n = -1;
 	}
 # else	/* !MSDOS */
-	if (statfs2(path, &fsbuf) < 0) return(-1);
-	*totalp = fsbuf.f_blocks;
-	*freep = fsbuf.f_bavail;
-	*bsizep = blocksize(fsbuf);
+	if (statfs2(path, &fsbuf) < 0) n = -1;
+	else {
+		*totalp = fsbuf.f_blocks;
+		*freep = fsbuf.f_bavail;
+		*bsizep = blocksize(fsbuf);
+	}
 # endif	/* !MSDOS */
 
-	return(0);
+	if (n < 0) {
+#if	MSDOS
+		*bsizep = (off_t)1024;
+#else
+		if (Xstat(path, &st) < 0) *bsizep = (off_t)DEV_BSIZE;
+		else *bsizep = (off_t)(st.st_blksize);
+#endif
+	}
+
+	return(n);
 }
 
 # if	MSDOS
 /*ARGSUSED*/
 static char *NEAR realpath2(path, resolved, rdlink)
-char *path, *resolved;
+CONST char *path;
+char *resolved;
 int rdlink;
 {
 	struct SREGS sreg;
@@ -573,8 +632,8 @@ time_t tim;
 
 #   ifdef	USEUTIME
 static int NEAR Xutime(path, times)
-char *path;
-struct utimbuf *times;
+CONST char *path;
+CONST struct utimbuf *times;
 {
 	time_t t;
 	__dpmi_regs reg;
@@ -584,8 +643,8 @@ struct utimbuf *times;
 	t = times -> modtime;
 #   else	/* !USEUTIME */
 static int NEAR Xutimes(path, tvp)
-char *path;
-struct timeval tvp[2];
+CONST char *path;
+CONST struct timeval *tvp;
 {
 	time_t t;
 	__dpmi_regs reg;
@@ -634,8 +693,8 @@ int on;
 }
 
 /*ARGSUSED*/
-static int NEAR getkey2(sig)
-int sig;
+static int NEAR getkey3(sig, code)
+int sig, code;
 {
 	u_char ch;
 	int i;
@@ -648,7 +707,7 @@ int sig;
 # endif	/* !MSDOS */
 
 static int NEAR touchfile(path, stp)
-char *path;
+CONST char *path;
 struct stat *stp;
 {
 # ifdef	USEUTIME
@@ -672,7 +731,7 @@ struct stat *stp;
 
 # ifndef	NODIRLOOP
 static int NEAR issamebody(src, dest)
-char *src, *dest;
+CONST char *src, *dest;
 {
 	struct stat st1, st2;
 
@@ -684,7 +743,7 @@ char *src, *dest;
 
 # ifndef	NOSYMLINK
 static int NEAR cpsymlink(src, dest)
-char *src, *dest;
+CONST char *src, *dest;
 {
 	struct stat st;
 	char path[MAXPATHLEN];
@@ -705,9 +764,9 @@ char *src, *dest;
 
 /*ARGSUSED*/
 static char *NEAR inputstr(prompt, delsp, ptr, def, h)
-char *prompt;
+CONST char *prompt;
 int delsp, ptr;
-char *def;
+CONST char *def;
 int h;
 {
 	char *cp;
@@ -747,7 +806,7 @@ int h;
 #endif	/* !FD */
 
 static VOID NEAR doserror(s, n)
-char *s;
+CONST char *s;
 int n;
 {
 	if (!n || n >= DOSERRSIZ) return;
@@ -760,7 +819,7 @@ int n;
 }
 
 static VOID NEAR dosperror(s)
-char *s;
+CONST char *s;
 {
 	int duperrno;
 
@@ -797,7 +856,7 @@ static int NEAR inputkey(VOID_A)
 	int c;
 
 	ttyiomode(1);
-	c = getkey2(0);
+	c = getkey3(0, inputkcode);
 	stdiomode();
 	if (c == EOF) c = -1;
 	else if (c == cc_intr) {
@@ -846,7 +905,9 @@ CONST VOID_P vp2;
 				else ret = (cp2) ? -1 : 0;
 				break;
 			case 'D':
-				ret = sp1 -> mtim - sp2 -> mtim;
+				if (sp1 -> mtim == sp2 -> mtim) ret = 0;
+				else ret = (sp1 -> mtim > sp2 -> mtim)
+					? 1 : -1;
 				break;
 			case 'G':
 				ret = dir_isdir(sp2) - dir_isdir(sp1);
@@ -869,8 +930,8 @@ CONST VOID_P vp2;
 }
 
 static VOID NEAR evalenvopt(cmd, env, getoptcmd)
-char *cmd, *env;
-int (NEAR *getoptcmd)__P_((int, char *[]));
+CONST char *cmd, *env;
+int (NEAR *getoptcmd)__P_((int, char *CONST []));
 {
 	char *cp, **argv;
 	int n, er, argc;
@@ -895,7 +956,7 @@ int (NEAR *getoptcmd)__P_((int, char *[]));
 
 static int NEAR getdiropt(argc, argv)
 int argc;
-char *argv[];
+char *CONST argv[];
 {
 	char *arg;
 	int i, j, n, r, rr;
@@ -1197,15 +1258,17 @@ struct filestat_t *dirp;
 static VOID NEAR showfnameb(dirp)
 struct filestat_t *dirp;
 {
-	char buf[MAXPATHLEN];
+	char *cp, buf[MAXPATHLEN];
 	int i;
 
 	if (dirflag & DF_SUBDIR) {
+		cp = dirwd;
 		if (dirflag & DF_LOWER) {
 			for (i = 0; dirwd[i]; i++) buf[i] = tolower2(dirwd[i]);
-			dirwd = buf;
+			buf[i] = '\0';
+			cp = buf;
 		}
-		fprintf2(stdout, "%k%c", dirwd, _SC_);
+		fprintf2(stdout, "%k%c", cp, _SC_);
 	}
 
 	if (dirflag & DF_LOWER)
@@ -1515,15 +1578,18 @@ off_t *sump, *bsump;
 
 int doscomdir(argc, argv)
 int argc;
-char *argv[];
+char *CONST argv[];
 {
 #ifndef	MINIMUMSHELL
 	off_t bsum;
+#endif
+#if	defined (FD) && !defined (_NODOSDRIVE)
+	int drv;
 #endif
 	reg_t *re;
 	struct stat st;
-	char *dir, *file;
-	char wd[MAXPATHLEN], cwd[MAXPATHLEN], buf[MAXPATHLEN];
+	CONST char *cp, *dir;
+	char *file, wd[MAXPATHLEN], cwd[MAXPATHLEN], buf[MAXPATHLEN];
 	off_t sum, total, fre, bsize;
 	int i, n, nf, nd;
 
@@ -1531,8 +1597,12 @@ char *argv[];
 	dirsort[0] = dirtype = '\0';
 	dirflag = 0;
 	evalenvopt(argv[0], "DIRCMD", getdiropt);
-	if ((n = getdiropt(argc, argv)) < 0)
-		return(RET_FAIL);
+	if ((n = getdiropt(argc, argv)) < 0) return(RET_FAIL);
+
+#if	defined (FD) && !defined (_NODOSDRIVE)
+	drv = -1;
+#endif
+
 	if (argc <= n) {
 		dir = curpath;
 		file = "*";
@@ -1541,46 +1611,52 @@ char *argv[];
 		doserror(argv[n + 1], ER_TOOMANYPARAM);
 		return(RET_FAIL);
 	}
-	else if (Xstat(argv[n], &st) >= 0
-	&& (st.st_mode & S_IFMT) == S_IFDIR) {
-		strcpy(buf, argv[n]);
-		dir = buf;
-		file = "*";
-	}
 	else {
 		strcpy(buf, argv[n]);
-		if (!(file = strrdelim(buf, 1))) {
-			dir = curpath;
-			file = argv[n];
+#if	defined (FD) && !defined (_NODOSDRIVE)
+		if (preparedrv(dospath3(buf), &drv) < 0) {
+			dosperror(buf);
+			return(RET_FAIL);
 		}
-		else {
+#endif
+		if (Xstat(buf, &st) >= 0
+		&& (st.st_mode & S_IFMT) == S_IFDIR) {
+			dir = buf;
+			file = "*";
+		}
+		else if ((file = strrdelim(buf, 1))) {
 			i = file - buf;
-			if (isrootdir(buf) || *file != _SC_) file++;
+			if (isrootpath(buf) || *file != _SC_) file++;
 			*file = '\0';
 			dir = buf;
 			file = &(argv[n][++i]);
 			if (!*file) file = "*";
 		}
+		else {
+			dir = curpath;
+			file = argv[n];
+		}
 	}
 
 	if (!Xgetwd(cwd)) {
+#if	defined (FD) && !defined (_NODOSDRIVE)
+		shutdrv(drv);
+#endif
 		dosperror(NULL);
 		return(RET_FAIL);
 	}
+
 	if (dir != buf) strcpy(wd, cwd);
+	else if ((cp = getrealpath(buf, wd, cwd)) == wd) /*EMPTY*/;
+#ifdef	DOUBLESLASH
+	else if (cp == buf && isdslash(buf)) realpath2(buf, wd, 1);
+#endif
 	else {
-		if (chdir2(buf) < 0) {
-			dosperror(buf);
-			return(RET_FAIL);
-		}
-		if (!Xgetwd(wd)) {
-			dosperror(NULL);
-			return(RET_FAIL);
-		}
-		if (chdir2(cwd) < 0) {
-			dosperror(cwd);
-			return(RET_FAIL);
-		}
+#if	defined (FD) && !defined (_NODOSDRIVE)
+		shutdrv(drv);
+#endif
+		dosperror(cp);
+		return(RET_FAIL);
 	}
 	if (getinfofs(wd, &total, &fre, &bsize) < 0) total = fre = (off_t)-1;
 
@@ -1597,6 +1673,9 @@ char *argv[];
 	n = dosdir(re, &bsize, &nf, &nd, &sum);
 #else
 	n = dosdir(re, &bsize, &nf, &nd, &sum, &bsum);
+#endif
+#if	defined (FD) && !defined (_NODOSDRIVE)
+		shutdrv(drv);
 #endif
 	if (re) regexp_free(re);
 	if (n < 0) {
@@ -1617,7 +1696,7 @@ char *argv[];
 
 static int NEAR checkarg(argc, argv)
 int argc;
-char *argv[];
+char *CONST argv[];
 {
 	if (argc <= 1) {
 		doserror(NULL, ER_REQPARAM);
@@ -1633,7 +1712,7 @@ char *argv[];
 
 int doscommkdir(argc, argv)
 int argc;
-char *argv[];
+char *CONST argv[];
 {
 	if (checkarg(argc, argv) < 0) return(RET_FAIL);
 
@@ -1647,7 +1726,7 @@ char *argv[];
 
 int doscomrmdir(argc, argv)
 int argc;
-char *argv[];
+char *CONST argv[];
 {
 	if (checkarg(argc, argv) < 0) return(RET_FAIL);
 
@@ -1661,10 +1740,10 @@ char *argv[];
 
 int doscomerase(argc, argv)
 int argc;
-char *argv[];
+char *CONST argv[];
 {
 	char *buf, **wild;
-	int i, n, key, flag, ret;
+	int i, n, c, flag, ret;
 
 	flag = 0;
 	for (n = 1; n < argc; n++) {
@@ -1702,10 +1781,10 @@ char *argv[];
 			stdiomode();
 			if (!buf) return(RET_SUCCESS);
 			if (!isatty(STDIN_FILENO)) fputnl(stdout);
-			key = *buf;
+			c = *buf;
 			free(buf);
-		} while (!strchr("ynYN", key));
-		if (key == 'n' || key == 'N') return(RET_SUCCESS);
+		} while (!strchr("ynYN", c));
+		if (c == 'n' || c == 'N') return(RET_SUCCESS);
 	}
 	if (!(wild = evalwild(argv[n], 0))) {
 		doserror(argv[n], ER_FILENOTFOUND);
@@ -1718,16 +1797,15 @@ char *argv[];
 				fprintf2(stdout,
 					"%k,    Delete (Y/N)?", wild[i]);
 				fflush(stdout);
-				if ((key = inputkey()) < 0) {
+				if ((c = inputkey()) < 0) {
 					freevar(wild);
 					return(ret);
 				}
-				if (key <= (int)MAXUTYPE(u_char)
-				&& isprint2(key))
-					fputc(key, stdout);
+				if (c <= (int)MAXUTYPE(u_char) && isprint2(c))
+					fputc(c, stdout);
 				fputnl(stdout);
-			} while (!strchr("ynYN", key));
-			if (key == 'n' || key == 'N') continue;
+			} while (!strchr("ynYN", c));
+			if (c == 'n' || c == 'N') continue;
 		}
 		if (Xunlink(wild[i]) < 0) {
 			dosperror(wild[i]);
@@ -1741,7 +1819,8 @@ char *argv[];
 }
 
 static char *NEAR convwild(dest, src, wild, swild)
-char *dest, *src, *wild, *swild;
+char *dest;
+CONST char *src, *wild, *swild;
 {
 	int i, j, n, rest, w;
 
@@ -1768,7 +1847,7 @@ char *dest, *src, *wild, *swild;
 
 int doscomrename(argc, argv)
 int argc;
-char *argv[];
+char *CONST argv[];
 {
 	char *cp, **wild, new[MAXPATHLEN];
 	int i, j, ret;
@@ -1817,7 +1896,7 @@ char *argv[];
 
 static int NEAR getcopyopt(argc, argv)
 int argc;
-char *argv[];
+char *CONST argv[];
 {
 	char *arg;
 	int i, j;
@@ -1864,27 +1943,21 @@ char *argv[];
 }
 
 static int NEAR getbinmode(name, bin)
-char *name;
+CONST char *name;
 int bin;
 {
 	char *cp;
 
 	if (bin < 0) return(-1);
-	if ((cp = strchr(name, DOSCOMOPT))) {
-		if (!cp[1] || (cp[2] && cp[3])) {
-			doserror(cp, ER_INVALIDSW);
-			return(-1);
-		}
-		*(cp++) = '\0';
-		if (toupper2(*cp) == 'B') bin = CF_BINARY;
-		else if (toupper2(*cp) == 'A') bin = CF_TEXT;
-	}
+	if (!(cp = strchr(name, DOSCOMOPT)) || !cp[1] || cp[2]) /*EMPTY*/;
+	else if (toupper2(*cp) == 'B') bin = CF_BINARY;
+	else if (toupper2(*cp) == 'A') bin = CF_TEXT;
 
 	return(bin);
 }
 
 static int NEAR writeopen(file, src)
-char *file, *src;
+CONST char *file, *src;
 {
 	struct stat st;
 	char buf[MAXPATHLEN], buf2[MAXPATHLEN];
@@ -1921,7 +1994,7 @@ char *file, *src;
 		if (key == 'N') return(0);
 		else if (key == 'A') copyflag |= CF_NOCONFIRM;
 	}
-	else if (copyflag & CF_VERBOSE) {
+	else if (src && (copyflag & CF_VERBOSE)) {
 		kanjifputs(src, stdout);
 		fputnl(stdout);
 	}
@@ -1930,7 +2003,7 @@ char *file, *src;
 	flags |= (O_BINARY | O_CREAT | O_TRUNC);
 
 #ifndef	NODIRLOOP
-	if (issamebody(src, file)) {
+	if (src && issamebody(src, file)) {
 		flags |= O_EXCL;
 		if (Xunlink(file) < 0) return(-1);
 	}
@@ -1986,7 +2059,7 @@ int fd, bin;
 }
 
 static int NEAR doscopy(src, dest, stp, sbin, dbin, dfd)
-char *src, *dest;
+CONST char *src, *dest;
 struct stat *stp;
 int sbin, dbin, dfd;
 {
@@ -2079,16 +2152,18 @@ int sbin, dbin, dfd;
 	}
 
 #ifdef	FD
-	stp -> st_nlink = (TCH_ATIME | TCH_MTIME);
-#endif
+	stp -> st_nlink = (TCH_ATIME | TCH_MTIME | TCH_IGNOREERR);
 	if (!tty && touchfile(dest, stp) < 0) return(-1);
+#else
+	if (!tty) VOID_C touchfile(dest, stp);
+#endif
 
 	return(1);
 }
 
 int doscomcopy(argc, argv)
 int argc;
-char *argv[];
+char *CONST argv[];
 {
 	struct stat sst, dst;
 	char *cp, *file, *form, **arg, *src, **wild;
@@ -2255,7 +2330,7 @@ char *argv[];
 /*ARGSUSED*/
 int doscomcls(argc, argv)
 int argc;
-char *argv[];
+char *CONST argv[];
 {
 #ifdef	FD
 	putterms(T_CLEAR);
@@ -2271,7 +2346,7 @@ char *argv[];
 
 int doscomtype(argc, argv)
 int argc;
-char *argv[];
+char *CONST argv[];
 {
 	char *cp;
 	u_char ch;
