@@ -13,12 +13,6 @@
 #include <process.h>
 # ifndef	DJGPP
 # include <dos.h>
-#  ifdef	__TURBOC__
-extern u_int _stklen = 0x5800;
-#  define	harderr_t	void
-#  else
-#  define	harderr_t	int
-#  endif
 # endif
 #endif	/* !MSDOS */
 
@@ -32,6 +26,17 @@ extern u_int _stklen = 0x5800;
 #define	isorgpid()	(mypid == orgpid)
 #endif
 
+#if	!defined (_NOIME) && defined (DEBUG)
+#include "roman.h"
+#endif
+
+#ifdef	__TURBOC__
+extern u_int _stklen = 0x5800;
+#define	harderr_t	void
+#else
+#define	harderr_t	int
+#endif
+
 #ifndef	DATADIR
 #define	DATADIR		progpath
 #endif
@@ -62,7 +67,7 @@ extern int maxarchive;
 #endif
 extern int wheader;
 extern char fullpath[];
-extern int savehist;
+extern short savehist[];
 extern char *helpindex[];
 extern int subwindow;
 extern int win_x;
@@ -74,6 +79,9 @@ extern char *deftmpdir;
 #ifdef	_USEUNICODE
 extern char *unitblpath;
 #endif
+#ifndef	_NOIME
+extern char *dicttblpath;
+#endif
 #ifndef	_NOPTY
 extern int ptymode;
 extern int ptyinternal;
@@ -151,20 +159,23 @@ static VOID NEAR printtime __P_((int));
 #ifdef	SIGALRM
 static int trapalrm __P_((VOID_A));
 #endif
+#ifndef	_NODOSDRIVE
+static int wrap_intrkey __P_((VOID_A));
+#endif
 static char *NEAR getversion __P_((int *));
 #ifndef	_NOLOGGING
-static VOID NEAR startlog __P_((char **));
+static VOID NEAR startlog __P_((char *CONST *));
 static VOID NEAR endlog __P_((int));
 #endif
 #ifdef	_NOORIGSHELL
-static int NEAR execruncomline __P_((char *, char *, int, char *));
+static int NEAR execruncomline __P_((char *, CONST char *, int, CONST char *));
 #endif
-static int NEAR initoption __P_((int, char *[], char *[]));
-static int NEAR evaloption __P_((char *[]));
-static char *NEAR searchenv __P_((char *, char *[]));
-static VOID NEAR setexecname __P_((char *));
-static VOID NEAR setexecpath __P_((char *, char *[]));
-int main __P_((int, char *[], char *[]));
+static int NEAR initoption __P_((int, char *CONST []));
+static int NEAR evaloption __P_((char *CONST []));
+static char *NEAR searchenv __P_((CONST char *, char *CONST []));
+static VOID NEAR setexecname __P_((CONST char *));
+static VOID NEAR setexecpath __P_((CONST char *, char *CONST []));
+int main __P_((int, char *CONST [], char *CONST []));
 
 char *origpath = NULL;
 char *progpath = NULL;
@@ -221,11 +232,16 @@ u_short far *devhdr;
 #endif
 
 VOID error(s)
-char *s;
+CONST char *s;
 {
+	static int doing = 0;
 	int duperrno;
 
+	if (doing == 1) exit2(2);
+	if (doing == 2) exit(2);
 	duperrno = errno;
+	doing = 1;
+
 	sigvecset(0);
 	if (isorgpid()) {
 		forcecleandir(deftmpdir, tmpfilename);
@@ -244,6 +260,8 @@ char *s;
 		fputs(s, stderr);
 		fputnl(stderr);
 	}
+	fclose(stderr);
+	doing = 2;
 
 	if (isorgpid()) {
 		inittty(1);
@@ -500,13 +518,13 @@ VOID setlinecol(VOID_A)
 {
 	char buf[MAXLONGWIDTH + 1];
 
-	if (getconstvar("LINES")) {
+	if (getconstvar(ENVLINES)) {
 		snprintf2(buf, sizeof(buf), "%d", n_line);
-		setenv2("LINES", buf, 1);
+		setenv2(ENVLINES, buf, 1);
 	}
-	if (getconstvar("COLUMNS")) {
+	if (getconstvar(ENVCOLUMNS)) {
 		snprintf2(buf, sizeof(buf), "%d", n_column);
-		setenv2("COLUMNS", buf, 1);
+		setenv2(ENVCOLUMNS, buf, 1);
 	}
 }
 
@@ -546,7 +564,8 @@ int n, max;
 VOID checkscreen(xmax, ymax)
 int xmax, ymax;
 {
-	char *cp, *tty;
+	CONST char *cp;
+	char *tty;
 	int i, row, wastty, dupn_line, dupdumbterm;
 
 #ifdef	SIGWINCH
@@ -597,7 +616,7 @@ int xmax, ymax;
 		Xputterm(T_BELL);
 		Xcputnl();
 		Xtflush();
-		if (kbhit2(1000000L) && getkey2(0) == K_ESC) {
+		if (kbhit2(1000000L) && getkey3(0, inputkcode) == K_ESC) {
 			errno = 0;
 			error(INTR_K);
 		}
@@ -729,6 +748,13 @@ static int trapalrm(VOID_A)
 }
 #endif	/* SIGALRM */
 
+#ifndef	_NODOSDRIVE
+static int wrap_intrkey(VOID_A)
+{
+	return(intrkey(-1));
+}
+#endif	/* !_NODOSDRIVE */
+
 int sigvecset(set)
 int set;
 {
@@ -750,7 +776,7 @@ int set;
 #endif
 #ifndef	_NODOSDRIVE
 		doswaitfunc = waitmes;
-		dosintrfunc = intrkey;
+		dosintrfunc = wrap_intrkey;
 #endif
 		status = 1;
 	}
@@ -792,7 +818,7 @@ int *lenp;
 
 VOID title(VOID_A)
 {
-	char *cp;
+	CONST char *cp;
 	int i, len;
 
 	Xlocate(0, L_TITLE);
@@ -816,7 +842,7 @@ VOID title(VOID_A)
 		Xputch2('#');
 		i++;
 	}
-	cp = (iswellomit()) ? nullstr : " (c)1995-2006 T.Shirai  ";
+	cp = (iswellomit()) ? nullstr : " (c)1995-2007 T.Shirai  ";
 	Xcputs2(cp);
 	i = n_column - len - strlen2(cp) - i;
 	while (i-- > 0) Xputch2(' ');
@@ -847,7 +873,7 @@ VOID saveorigenviron(VOID_A)
 
 #ifndef	_NOLOGGING
 static VOID NEAR startlog(argv)
-char **argv;
+char *CONST *argv;
 {
 # ifndef	NOUID
 	uid_t uid;
@@ -902,9 +928,10 @@ int status;
 
 #ifdef	_NOORIGSHELL
 static int NEAR execruncomline(command, file, n, line)
-char *command, *file;
+char *command;
+CONST char *file;
 int n;
-char *line;
+CONST char *line;
 {
 	char *cp;
 	int i;
@@ -927,22 +954,23 @@ char *line;
 
 /*ARGSUSED*/
 int loadruncom(file, exist)
-char *file;
+CONST char *file;
 int exist;
 {
 #ifdef	_NOORIGSHELL
+# if	!MSDOS
+	struct stat st;
+# endif
 	FILE *fp;
 	char *cp, *tmp, *fold, *line;
 	int cont;
-#endif
+#endif	/* _NOORIGSHELL */
 	int n, er;
 
 #ifdef	_NOORIGSHELL
 # if	!MSDOS
 	tmp = NULL;
-	if (!exist && (tmp = getconstvar("TERM"))) {
-		struct stat st;
-
+	if (!exist && (tmp = getconstvar(ENVTERM))) {
 		cp = malloc2(strlen(file) + strlen(tmp) + 1 + 1);
 		strcpy(strcpy2(strcpy2(cp, file), "."), tmp);
 		tmp = evalpath(cp, 0);
@@ -976,7 +1004,7 @@ int exist;
 		}
 
 		cp = line + strlen(line);
-		for (cp--; cp >= line && (*cp == ' ' || *cp == '\t'); cp--);
+		for (cp--; cp >= line && isblank2(*cp); cp--);
 		cp[1] = '\0';
 
 		cont = 0;
@@ -1015,10 +1043,9 @@ int exist;
 	return(er ? -1 : 0);
 }
 
-/*ARGSUSED*/
-static int NEAR initoption(argc, argv, envp)
+static int NEAR initoption(argc, argv)
 int argc;
-char *argv[], *envp[];
+char *CONST argv[];
 {
 	char *cp, **optv;
 	int i, optc;
@@ -1048,14 +1075,14 @@ char *argv[], *envp[];
 
 #ifndef	_NOORIGSHELL
 	if (initshell(optc, optv) < 0) Xexit2(RET_FAIL);
-#endif	/* !_NOORIGSHELL */
+#endif
 	free(optv);
 
 	return(argc);
 }
 
 static int NEAR evaloption(argv)
-char *argv[];
+char *CONST argv[];
 {
 	char *cp, *tmp;
 	int i;
@@ -1076,7 +1103,8 @@ char *argv[];
 }
 
 static char *NEAR searchenv(s, envp)
-char *s, *envp[];
+CONST char *s;
+char *CONST envp[];
 {
 	int i, len;
 
@@ -1089,7 +1117,7 @@ char *s, *envp[];
 }
 
 static VOID NEAR setexecname(argv)
-char *argv;
+CONST char *argv;
 {
 #if	MSDOS || defined (CYGWIN)
 	char *cp;
@@ -1105,13 +1133,15 @@ char *argv;
 }
 
 static VOID NEAR setexecpath(argv, envp)
-char *argv, *envp[];
+CONST char *argv;
+char *CONST envp[];
 {
-	char *cp, buf[MAXPATHLEN];
+	CONST char *cp;
+	char *tmp, buf[MAXPATHLEN];
 
 	if (!Xgetwd(buf)) error(NOCWD_K);
 	origpath = strdup2(buf);
-	if ((cp = searchenv("PWD", envp))) {
+	if ((cp = searchenv(ENVPWD, envp))) {
 		*fullpath = '\0';
 		realpath2(cp, fullpath, 0);
 		realpath2(fullpath, buf, 1);
@@ -1122,26 +1152,30 @@ char *argv, *envp[];
 	}
 	strcpy(fullpath, origpath);
 
+	tmp = NULL;
 #if	MSDOS
 	cp = argv;
 #else
 	if (strdelim(argv, 0)) cp = argv;
-	else if ((cp = searchenv("PATH", envp))) cp = searchexecpath(argv, cp);
+	else if ((cp = searchenv(ENVPATH, envp)))
+		cp = tmp = searchexecpath(argv, cp);
 	if (!cp) progpath = strdup2(origpath);
 	else
 #endif
 	{
 		realpath2(cp, buf, 1);
-		if (cp != argv) free(cp);
-		if ((cp = strrdelim(buf, 0))) *cp = '\0';
+		if (tmp) free(tmp);
+		if ((tmp = strrdelim(buf, 0))) *tmp = '\0';
 		progpath = strdup2(buf);
 	}
 }
 
 /*ARGSUSED*/
 VOID initfd(argv)
-char **argv;
+char *CONST *argv;
 {
+	int i;
+
 #ifndef	_NOORIGSHELL
 	if (!interactive) /*EMPTY*/;
 	else
@@ -1154,7 +1188,7 @@ char **argv;
 			Xttyiomode(0);
 		}
 #endif	/* !MSDOS */
-		loadhistory(0);
+		for (i = 0; i < 2; i++) loadhistory(i);
 		entryhist(1, origpath, 1);
 	}
 #ifndef	_NOLOGGING
@@ -1167,6 +1201,7 @@ VOID prepareexitfd(status)
 int status;
 {
 	char cwd[MAXPATHLEN];
+	int i;
 
 #ifndef	_NOLOGGING
 	endlog(status);
@@ -1175,7 +1210,7 @@ int status;
 	if (!interactive) /*EMPTY*/;
 	else
 #endif
-	savehistory(0);
+	for (i = 0; i < 2; i++) savehistory(i);
 
 	cwd[0] = '\0';
 	if (origpath && _chdir2(origpath) < 0) {
@@ -1188,6 +1223,9 @@ int status;
 #ifdef	_USEUNICODE
 	free(unitblpath);
 #endif
+#ifndef	_NOIME
+	free(dicttblpath);
+#endif
 #ifndef	_NODOSDRIVE
 	dosallclose();
 #endif
@@ -1242,13 +1280,18 @@ int status;
 # ifdef	_USEUNICODE
 	discardunitable();
 # endif
+# ifndef	_NOIME
+	ime_freebuf();
+	freeroman(0);
+	discarddicttable();
+# endif
 #endif	/* DEBUG */
 	if (*cwd) rawchdir(cwd);
 }
 
 int main(argc, argv, envp)
 int argc;
-char *argv[], *envp[];
+char *CONST argv[], *CONST envp[];
 {
 	char *cp;
 	int i;
@@ -1352,7 +1395,7 @@ char *argv[], *envp[];
 		fdtype[i].offset = 0;
 # endif
 	}
-#endif
+#endif	/* _USEDOSEMU */
 
 #ifdef	_NOORIGSHELL
 	for (i = 0; envp[i]; i++);
@@ -1361,16 +1404,18 @@ char *argv[], *envp[];
 	environ[i] = NULL;
 #endif
 
-	initenv();
 	setexecpath(argv[0], envp);
 #ifdef	_USEUNICODE
 	unitblpath = strdup2(DATADIR);
 #endif
+#ifndef	_NOIME
+	dicttblpath = strdup2(DATADIR);
+#endif
 
 #ifdef	_NOORIGSHELL
 	inittty(0);
 	getterment(NULL);
-	argc = initoption(argc, argv, envp);
+	argc = initoption(argc, argv);
 	adjustpath();
 #else	/* !_NOORIGSHELL */
 	cp = getshellname(progname, NULL, NULL);
@@ -1381,18 +1426,16 @@ char *argv[], *envp[];
 # endif
 		Xexit2(i);
 	}
-#ifndef	_NOORIGSHELL
-	interactive =
-#endif
-	fdmode = 1;
+	interactive = fdmode = 1;
 	setshellvar(envp);
 	prepareterm();
-	argc = initoption(argc, argv, envp);
+	argc = initoption(argc, argv);
 	if (dumbterm > 1) {
 		errno = 0;
 		error(NTERM_K);
 	}
 #endif	/* !_NOORIGSHELL */
+	initenv();
 
 	Xttyiomode(0);
 	initterm();
