@@ -80,53 +80,67 @@ extern char *_mtrace_file;
 #endif
 
 #ifdef	FD
-extern char *getenv2 __P_((char *));
+extern char *getenv2 __P_((CONST char *));
 extern char *malloc2 __P_((ALLOC_T));
 extern char *realloc2 __P_((VOID_P, ALLOC_T));
 extern char *c_realloc __P_((char *, ALLOC_T, ALLOC_T *));
-extern char *strdup2 __P_((char *));
-extern char *strndup2 __P_((char *, int));
-extern char *strchr2 __P_((char *, int));
-extern char *strcpy2 __P_((char *, char *));
-extern char *strncpy2 __P_((char *, char *, int));
+extern char *strdup2 __P_((CONST char *));
+extern char *strndup2 __P_((CONST char *, int));
+extern char *strchr2 __P_((CONST char *, int));
+extern char *strcpy2 __P_((char *, CONST char *));
+extern char *strncpy2 __P_((char *, CONST char *, int));
 # ifdef	_USEDOSPATH
-extern int _dospath __P_((char *));
+extern int _dospath __P_((CONST char *));
 # endif
-extern DIR *Xopendir __P_((char *));
+# if	MSDOS
+extern int dospath3 __P_((CONST char *));
+# else
+extern int dospath __P_((CONST char *, char *));
+#define	dospath3(path)	dospath(path, NULL)
+# endif
+extern DIR *Xopendir __P_((CONST char *));
 extern int Xclosedir __P_((DIR *));
 extern struct dirent *Xreaddir __P_((DIR *));
 extern char *Xgetwd __P_((char *));
-extern int Xstat __P_((char *, struct stat *));
-extern int stat2 __P_((char *, struct stat *));
-extern int Xaccess __P_((char *, int));
+extern int Xstat __P_((CONST char *, struct stat *));
+extern int stat2 __P_((CONST char *, struct stat *));
+extern int _chdir2 __P_((CONST char *));
+extern int Xaccess __P_((CONST char *, int));
+# ifndef	_NODOSDRIVE
+extern int preparedrv __P_((int, int *));
+extern VOID shutdrv __P_((int));
+# endif
 # if	MSDOS && !defined (_NOUSELFN)
-extern char *shortname __P_((char *, char *));
+extern char *shortname __P_((CONST char *, char *));
 # endif
 extern VOID demacroarg __P_((char **));
+extern VOID lostcwd __P_((char *));
 extern char *progpath;
 #else	/* !FD */
 #define	getenv2		(char *)getenv
-VOID error __P_((char *));
+VOID error __P_((CONST char *));
 char *malloc2 __P_((ALLOC_T));
 char *realloc2 __P_((VOID_P, ALLOC_T));
 char *c_realloc __P_((char *, ALLOC_T, ALLOC_T *));
-char *strdup2 __P_((char *));
-char *strndup2 __P_((char *, int));
-char *strchr2 __P_((char *, int));
-char *strcpy2 __P_((char *, char *));
-char *strncpy2 __P_((char *, char *, int));
+char *strdup2 __P_((CONST char *));
+char *strndup2 __P_((CONST char *, int));
+char *strchr2 __P_((CONST char *, int));
+char *strcpy2 __P_((char *, CONST char *));
+char *strncpy2 __P_((char *, CONST char *, int));
 # if	MSDOS
-int _dospath __P_((char *));
-DIR *Xopendir __P_((char *));
+int _dospath __P_((CONST char *));
+extern int chdir2 __P_((CONST char *));
+DIR *Xopendir __P_((CONST char *));
 int Xclosedir __P_((DIR *));
 struct dirent *Xreaddir __P_((DIR *));
 static u_int NEAR getdosmode __P_((u_int));
 static time_t NEAR getdostime __P_((u_int, u_int));
-int Xstat __P_((char *, struct stat *));
+int Xstat __P_((CONST char *, struct stat *));
 #  ifndef	NOSYMLINK
 #  define	Xlstat		Xstat
 #  endif
 # else	/* !MSDOS */
+# define	chdir2(p)	((chdir(p)) ? -1 : 0)
 # define	Xopendir	opendir
 # define	Xclosedir	closedir
 # define	Xreaddir	readdir
@@ -135,10 +149,11 @@ int Xstat __P_((char *, struct stat *));
 #  define	Xlstat(p, s)	((lstat(p, s)) ? -1 : 0)
 #  endif
 # endif	/* !MSDOS */
+# define	_chdir2		chdir2
 # ifdef	NOSYMLINK
 # define	stat2		Xstat
 # else
-int stat2 __P_((char *, struct stat *));
+int stat2 __P_((CONST char *, struct stat *));
 # endif
 # ifdef	DJGPP
 char *Xgetwd __P_((char *));
@@ -152,49 +167,52 @@ char *Xgetwd __P_((char *));
 #define	Xaccess(p, m)	((access(p, m)) ? -1 : 0)
 #endif	/* !FD */
 
-static char *NEAR getenvvar __P_((char *, int));
-static int NEAR setvar __P_((char *, char *, int));
-static int NEAR isescape __P_((char *s, int, int, int, int));
+static char *NEAR getenvvar __P_((CONST char *, int));
+static int NEAR setvar __P_((CONST char *, CONST char *, int));
+static int NEAR isescape __P_((CONST char *, int, int, int, int));
 #ifdef	_NOORIGGLOB
-static char *NEAR cnvregexp __P_((char *, int));
+static char *NEAR cnvregexp __P_((CONST char *, int));
 #else
-static int NEAR _regexp_exec __P_((char **, char *));
+static int NEAR _regexp_exec __P_((CONST reg_t *, CONST char *));
 #endif
-static void NEAR addstrbuf __P_((strbuf_t *, char *, int));
-static void NEAR duplwild __P_((wild_t *, wild_t *));
+static void NEAR addstrbuf __P_((strbuf_t *, CONST char *, int));
+static void NEAR duplwild __P_((wild_t *, CONST wild_t *));
 static void NEAR freewild __P_((wild_t *));
 static int NEAR _evalwild __P_((int, char ***, wild_t *));
 #ifndef	_NOUSEHASH
-static int NEAR calchash __P_((char *));
+static int NEAR calchash __P_((CONST char *));
 static VOID NEAR inithash __P_((VOID_A));
-static hashlist *NEAR newhash __P_((char *, char *, int, hashlist *));
+static hashlist *NEAR newhash __P_((CONST char *, char *, int, hashlist *));
 static VOID NEAR freehash __P_((hashlist **));
-static hashlist *NEAR findhash __P_((char *, int));
-static VOID NEAR rmhash __P_((char *, int));
+static hashlist *NEAR findhash __P_((CONST char *, int));
+static VOID NEAR rmhash __P_((CONST char *, int));
 #endif
-static int NEAR isexecute __P_((char *, int, int));
+static int NEAR isexecute __P_((CONST char *, int, int));
 #if	MSDOS
-static int NEAR extaccess __P_((char *, char *, int, int));
+static int NEAR extaccess __P_((char *, CONST char *, int, int));
 #endif
 #if	!defined (FDSH) && !defined (_NOCOMPLETE)
-static int NEAR completefile __P_((char *, int, int, char ***,
-	char *, int, int));
-static int NEAR completeexe __P_((char *, int, int, char ***));
+static int NEAR completefile __P_((CONST char *, int, int, char ***,
+		CONST char *, int, int));
+static int NEAR completeexe __P_((CONST char *, int, int, char ***));
 #endif	/* !FDSH && !_NOCOMPLETE */
-static int NEAR addmeta __P_((char *, char *, int));
-static int NEAR skipvar __P_((char **, int *, int *, int));
+static int NEAR addmeta __P_((char *, CONST char *, int));
+static int NEAR skipvar __P_((CONST char **, int *, int *, int));
 #ifdef	MINIMUMSHELL
-static int NEAR skipvarvalue __P_((char *, int *, char *, int, int));
+static int NEAR skipvarvalue __P_((CONST char *, int *,
+		CONST char *, int, int));
 static char *NEAR evalshellparam __P_((int, int));
 #else	/* !MINIMUMSHELL */
-static int NEAR skipvarvalue __P_((char *, int *, char *, int, int, int));
-static char *NEAR removeword __P_((char *, char *, int, int));
-static char **NEAR removevar __P_((char **, char *, int, int));
-static char *NEAR evalshellparam __P_((int, int, char *, int, int *));
+static int NEAR skipvarvalue __P_((CONST char *, int *,
+		CONST char *, int, int, int));
+static char *NEAR removeword __P_((CONST char *, CONST char *, int, int));
+static char **NEAR removevar __P_((char **, CONST char *, int, int));
+static char *NEAR evalshellparam __P_((int, int, CONST char *, int, int *));
 #endif	/* !MINIMUMSHELL */
-static int NEAR replacevar __P_((char *, char **, int, int, int, int, int));
-static char *NEAR insertarg __P_((char *, int, char *, int, int));
-static int NEAR evalvar __P_((char **, int, char **, int));
+static int NEAR replacevar __P_((CONST char *, char **,
+		int, int, int, int, int));
+static char *NEAR insertarg __P_((char *, int, CONST char *, int, int));
+static int NEAR evalvar __P_((char **, int, CONST char **, int));
 static char *NEAR replacebackquote __P_((char *, int *, char *, int));
 
 #define	BUFUNIT		32
@@ -203,26 +221,29 @@ static char *NEAR replacebackquote __P_((char *, int *, char *, int));
 			(((n) % BUFUNIT) ? ((type *)(ptr)) \
 			: (type *)realloc2(ptr, b_size(n, type)))
 #define	getconstvar(s)	(getenvvar(s, sizeof(s) - 1))
+#ifdef	FD
+#define	STRFD		"FD"
+#endif
 
-char nullstr[] = "";
-char rootpath[] = _SS_;
-char curpath[] = ".";
-char parentpath[] = "..";
+CONST char nullstr[] = "";
+CONST char rootpath[] = _SS_;
+CONST char curpath[] = ".";
+CONST char parentpath[] = "..";
 char **argvar = NULL;
 #ifndef	_NOUSEHASH
 hashlist **hashtable = NULL;
 #endif
-char *(*getvarfunc)__P_((char *, int)) = NULL;
+char *(*getvarfunc)__P_((CONST char *, int)) = NULL;
 int (*putvarfunc)__P_((char *, int)) = NULL;
 int (*getretvalfunc)__P_((VOID_A)) = NULL;
 p_id_t (*getpidfunc)__P_((VOID_A)) = NULL;
 p_id_t (*getlastpidfunc)__P_((VOID_A)) = NULL;
 char *(*getflagfunc)__P_((VOID_A)) = NULL;
-int (*checkundeffunc)__P_((char *, char *, int)) = NULL;
+int (*checkundeffunc)__P_((CONST char *, CONST char *, int)) = NULL;
 VOID (*exitfunc)__P_((VOID_A)) = NULL;
-char *(*backquotefunc)__P_((char *)) = NULL;
+char *(*backquotefunc)__P_((CONST char *)) = NULL;
 #ifndef	MINIMUMSHELL
-char *(*posixsubstfunc)__P_((char *, int *)) = NULL;
+char *(*posixsubstfunc)__P_((CONST char *, int *)) = NULL;
 #endif
 #ifndef	PATHNOCASE
 int pathignorecase = 0;
@@ -243,7 +264,7 @@ static int maxgid = 0;
 
 #ifndef	FD
 VOID error(s)
-char *s;
+CONST char *s;
 {
 	fputs(s, stderr);
 	fputs(": memory allocation error", stderr);
@@ -293,14 +314,13 @@ ALLOC_T n, *sizep;
 		*sizep = BUFUNIT;
 		return(malloc2(*sizep));
 	}
-	n++;
-	while (n >= *sizep) *sizep *= 2;
+	while (n + 1 >= *sizep) *sizep *= 2;
 
 	return(realloc2(ptr, *sizep));
 }
 
 char *strdup2(s)
-char *s;
+CONST char *s;
 {
 	char *tmp;
 	int n;
@@ -314,7 +334,7 @@ char *s;
 }
 
 char *strndup2(s, n)
-char *s;
+CONST char *s;
 int n;
 {
 	char *tmp;
@@ -330,13 +350,13 @@ int n;
 }
 
 char *strchr2(s, c)
-char *s;
+CONST char *s;
 int c;
 {
 	int i;
 
 	for (i = 0; s[i]; i++) {
-		if (s[i] == c) return(&(s[i]));
+		if (s[i] == c) return((char *)&(s[i]));
 		if (iskanji1(s, i)) i++;
 #ifdef	CODEEUC
 		else if (isekana(s, i)) i++;
@@ -347,7 +367,8 @@ int c;
 }
 
 char *strcpy2(s1, s2)
-char *s1, *s2;
+char *s1;
+CONST char *s2;
 {
 	int i;
 
@@ -358,7 +379,8 @@ char *s1, *s2;
 }
 
 char *strncpy2(s1, s2, n)
-char *s1, *s2;
+char *s1;
+CONST char *s2;
 int n;
 {
 	int i;
@@ -371,13 +393,13 @@ int n;
 
 # if	MSDOS
 int _dospath(path)
-char *path;
+CONST char *path;
 {
 	return((isalpha2(*path) && path[1] == ':') ? *path : 0);
 }
 
 DIR *Xopendir(dir)
-char *dir;
+CONST char *dir;
 {
 	DIR *dirp;
 	char *cp;
@@ -478,7 +500,7 @@ u_int d, t;
 }
 
 int Xstat(path, stp)
-char *path;
+CONST char *path;
 struct stat *stp;
 {
 	struct find_t find;
@@ -498,7 +520,7 @@ struct stat *stp;
 
 # ifndef	NOSYMLINK
 int stat2(path, stp)
-char *path;
+CONST char *path;
 struct stat *stp;
 {
 	int duperrno;
@@ -546,14 +568,14 @@ int drive, c;
 }
 
 char *strdelim(s, d)
-char *s;
+CONST char *s;
 int d;
 {
 	int i;
 
-	if (d && _dospath(s)) return(s + 1);
+	if (d && _dospath(s)) return((char *)&(s[1]));
 	for (i = 0; s[i]; i++) {
-		if (s[i] == _SC_) return(&(s[i]));
+		if (s[i] == _SC_) return((char *)&(s[i]));
 # ifdef	BSPATHDELIM
 		if (iskanji1(s, i)) i++;
 # endif
@@ -563,13 +585,13 @@ int d;
 }
 
 char *strrdelim(s, d)
-char *s;
+CONST char *s;
 int d;
 {
-	char *cp;
+	CONST char *cp;
 	int i;
 
-	if (d && _dospath(s)) cp = s + 1;
+	if (d && _dospath(s)) cp = &(s[1]);
 	else cp = NULL;
 	for (i = 0; s[i]; i++) {
 		if (s[i] == _SC_) cp = &(s[i]);
@@ -578,15 +600,16 @@ int d;
 # endif
 	}
 
-	return(cp);
+	return((char *)cp);
 }
 #endif	/* MSDOS || (FD && !_NODOSDRIVE) */
 
+#ifdef	FD
 char *strrdelim2(s, eol)
-char *s, *eol;
+CONST char *s, *eol;
 {
-#ifdef	BSPATHDELIM
-	char *cp;
+# ifdef	BSPATHDELIM
+	CONST char *cp;
 	int i;
 
 	cp = NULL;
@@ -595,16 +618,17 @@ char *s, *eol;
 		if (iskanji1(s, i)) i++;
 	}
 
-	return(cp);
-#else
-	for (eol--; eol >= s; eol--) if (*eol == _SC_) return(eol);
+	return((char *)cp);
+# else
+	for (eol--; eol >= s; eol--) if (*eol == _SC_) return((char *)eol);
 
 	return(NULL);
-#endif
+# endif
 }
+#endif	/* FD */
 
 int isdelim(s, ptr)
-char *s;
+CONST char *s;
 int ptr;
 {
 #ifdef	BSPATHDELIM
@@ -634,6 +658,10 @@ char *s;
 #if	MSDOS || (defined (FD) && !defined (_NODOSDRIVE))
 	if (_dospath(s)) i = 2;
 	else
+#endif
+#ifdef	DOUBLESLASH
+	if (isdslash(s)) i = 1;
+	else
 #endif
 	i = 0;
 	if (!s[i]) return(&(s[i]));
@@ -661,7 +689,8 @@ char *s;
 }
 
 char *strcatdelim2(buf, s1, s2)
-char *buf, *s1, *s2;
+char *buf;
+CONST char *s1, *s2;
 {
 	char *cp;
 	int i, len;
@@ -673,6 +702,13 @@ char *buf, *s1, *s2;
 		i = 2;
 	}
 	else
+#endif
+#ifdef	DOUBLESLASH
+	if (isdslash(s1)) {
+		buf[0] = s1[0];
+		i = 1;
+	}
+	else
 #endif
 	i = 0;
 	if (!s1[i]) cp = &(buf[i]);
@@ -683,6 +719,10 @@ char *buf, *s1, *s2;
 	else {
 		cp = NULL;
 		for (; s1[i]; i++) {
+			if (i >= MAXPATHLEN - 1) {
+				buf[i] = '\0';
+				return(&(buf[i]));
+			}
 			buf[i] = s1[i];
 			if (s1[i] == _SC_) {
 				if (!cp) cp = &(buf[i]) + 1;
@@ -715,15 +755,19 @@ char *buf, *s1, *s2;
 }
 
 int strcasecmp2(s1, s2)
-char *s1, *s2;
+CONST char *s1, *s2;
 {
+	int c1, c2;
+
 	for (;;) {
-		if (toupper2(*s1) != toupper2(*s2)) return(*s1 - *s2);
+		c1 = toupper2(*s1);
+		c2 = toupper2(*s2);
+		if (c1 != c2) return(c1 - c2);
 #ifndef	CODEEUC
 		if (issjis1(*s1)) {
 			s1++;
 			s2++;
-			if (*s1 != *s2) return(*s1 - *s2);
+			if (*s1 != *s2) return((u_char)*s1 - (u_char)*s2);
 		}
 #endif
 		if (!*s1) break;
@@ -735,17 +779,21 @@ char *s1, *s2;
 }
 
 int strncasecmp2(s1, s2, n)
-char *s1, *s2;
+CONST char *s1, *s2;
 int n;
 {
+	int c1, c2;
+
 	while (n-- > 0) {
-		if (toupper2(*s1) != toupper2(*s2)) return(*s1 - *s2);
+		c1 = toupper2(*s1);
+		c2 = toupper2(*s2);
+		if (c1 != c2) return(c1 - c2);
 #ifndef	CODEEUC
 		if (issjis1(*s1)) {
 			if (n-- <= 0) break;
 			s1++;
 			s2++;
-			if (*s1 != *s2) return(*s1 - *s2);
+			if (*s1 != *s2) return((u_char)*s1 - (u_char)*s2);
 		}
 #endif
 		if (!*s1) break;
@@ -758,7 +806,7 @@ int n;
 
 #ifndef	PATHNOCASE
 int strpathcmp2(s1, s2)
-char *s1, *s2;
+CONST char *s1, *s2;
 {
 	if (pathignorecase) return(strcasecmp2(s1, s2));
 
@@ -766,7 +814,7 @@ char *s1, *s2;
 }
 
 int strnpathcmp2(s1, s2, n)
-char *s1, *s2;
+CONST char *s1, *s2;
 int n;
 {
 	if (pathignorecase) return(strncasecmp2(s1, s2, n));
@@ -775,22 +823,24 @@ int n;
 }
 #endif	/* !PATHNOCASE */
 
+#ifdef	FD
 char *underpath(path, dir, len)
-char *path, *dir;
+CONST char *path, *dir;
 int len;
 {
 	char *cp;
 
 	if (len < 0) len = strlen(dir);
-	if ((cp = strrdelim2(dir, &(dir[len]))) && !*(cp + 1)) len = cp - dir;
+	if ((cp = strrdelim2(dir, &(dir[len]))) && !cp[1]) len = cp - dir;
 	if (len <= 0 || strnpathcmp(path, dir, len)) return(NULL);
 	if (path[len] && path[len] != _SC_) return(NULL);
 
-	return(&(path[len]));
+	return((char *)&(path[len]));
 }
+#endif	/* FD */
 
 static char *NEAR getenvvar(ident, len)
-char *ident;
+CONST char *ident;
 int len;
 {
 #if	(!defined (FD) && !defined (FDSH)) || defined (_NOORIGSHELL)
@@ -811,7 +861,7 @@ int len;
 }
 
 static int NEAR setvar(ident, value, len)
-char *ident, *value;
+CONST char *ident, *value;
 int len;
 {
 	char *cp;
@@ -846,7 +896,7 @@ int c;
 }
 
 int isdotdir(s)
-char *s;
+CONST char *s;
 {
 	if (s[0] != '.') /*EMPTY*/;
 	else if (!s[1]) return(2);
@@ -857,20 +907,20 @@ char *s;
 }
 
 char *isrootdir(s)
-char *s;
+CONST char *s;
 {
 #if	MSDOS || (defined (FD) && !defined (_NODOSDRIVE))
 	if (_dospath(s)) s += 2;
 #endif
 	if (*s != _SC_) return(NULL);
 
-	return(s);
+	return((char *)s);
 }
 
 int isrootpath(s)
-char *s;
+CONST char *s;
 {
-	return((s[0] == _SC_ && !s[1]));
+	return((s[0] == _SC_ && !s[1]) ? 1 : 0);
 }
 
 VOID copyrootpath(s)
@@ -887,18 +937,35 @@ char *s;
 	*s = '\0';
 }
 
+#ifdef	DOUBLESLASH
+int isdslash(s)
+CONST char *s;
+{
+# if	MSDOS || defined (CYGWIN)
+	char *cp;
+
+	if (s[0] != _SC_ || s[1] != _SC_ || !s[2] || s[2] == _SC_) return(0);
+	if ((cp = strdelim(&(s[2]), 0))) return(cp - s);
+
+	return(strlen(s));
+# else
+	return((s[0] == _SC_ && s[1] == _SC_) ? 2 : 0);
+# endif
+}
+#endif	/* DOUBLESLASH */
+
 char *getbasename(s)
-char *s;
+CONST char *s;
 {
 	char *cp;
 
-	if ((cp = strrdelim(s, 1))) return(cp + 1);
+	if ((cp = strrdelim(s, 1))) return(&(cp[1]));
 
-	return(s);
+	return((char *)s);
 }
 
 char *getshellname(s, loginp, restrictedp)
-char *s;
+CONST char *s;
 int *loginp, *restrictedp;
 {
 	if (loginp) *loginp = 0;
@@ -917,11 +984,11 @@ int *loginp, *restrictedp;
 		if (restrictedp) *restrictedp = 1;
 	}
 
-	return(s);
+	return((char *)s);
 }
 
 static int NEAR isescape(s, ptr, quote, len, flags)
-char *s;
+CONST char *s;
 int ptr, quote, len, flags;
 {
 #ifdef	FAKEMETA
@@ -961,7 +1028,7 @@ int ptr, quote, len, flags;
 
 #ifdef	_NOORIGGLOB
 static char *NEAR cnvregexp(s, len)
-char *s;
+CONST char *s;
 int len;
 {
 	char *re;
@@ -1026,24 +1093,26 @@ extern int re_ignore_case;
 # endif
 
 reg_t *regexp_init(s, len)
-char *s;
+CONST char *s;
 int len;
 {
+	char *new;
+
 # ifdef	PATHNOCASE
 	re_ignore_case = 1;
 # endif
 	skipdotfile = (*s == '*' || *s == '?' || *s == '[');
-	s = cnvregexp(s, len);
-	re_comp(s);
-	free(s);
+	new = cnvregexp(s, len);
+	re_comp(new);
+	free(new);
 
 	return((reg_t *)1);
 }
 
 /*ARGSUSED*/
 int regexp_exec(re, s, fname)
-reg_t *re;
-char *s;
+CONST reg_t *re;
+CONST char *s;
 int fname;
 {
 	if (fname && skipdotfile && *s == '.') return(0);
@@ -1062,13 +1131,14 @@ reg_t *re;
 # ifdef	USEREGCOMP
 
 reg_t *regexp_init(s, len)
-char *s;
+CONST char *s;
 int len;
 {
 	reg_t *re;
+	char *new;
 
 	skipdotfile = (*s == '*' || *s == '?' || *s == '[');
-	s = cnvregexp(s, len);
+	s = new = cnvregexp(s, len);
 	re = (reg_t *)malloc2(sizeof(reg_t));
 # ifdef	PATHNOCASE
 	if (regcomp(re, s, REG_EXTENDED | REG_ICASE)) {
@@ -1078,14 +1148,14 @@ int len;
 		free(re);
 		re = NULL;
 	}
-	free(s);
+	free(new);
 
 	return(re);
 }
 
 int regexp_exec(re, s, fname)
-reg_t *re;
-char *s;
+CONST reg_t *re;
+CONST char *s;
 int fname;
 {
 	if (!re || (fname && skipdotfile && *s == '.')) return(0);
@@ -1104,26 +1174,27 @@ reg_t *re;
 
 # else	/* !USEREGCOMP */
 #  ifdef	USEREGCMP
-extern char *regcmp __P_((char *, int));
-extern char *regex __P_((char *, char *));
+extern char *regcmp __P_((CONST char *, int));
+extern char *regex __P_((CONST char *, CONST char *, ...));
 
 reg_t *regexp_init(s, len)
-char *s;
+CONST char *s;
 int len;
 {
 	reg_t *re;
+	char *new;
 
 	skipdotfile = (*s == '*' || *s == '?' || *s == '[');
-	s = cnvregexp(s, len);
-	re = regcmp(s, 0);
-	free(s);
+	new = cnvregexp(s, len);
+	re = regcmp(new, 0);
+	free(new);
 
 	return(re);
 }
 
 int regexp_exec(re, s, fname)
-reg_t *re;
-char *s;
+CONST reg_t *re;
+CONST char *s;
 int fname;
 {
 	if (!re || (fname && skipdotfile && *s == '.')) return(0);
@@ -1139,7 +1210,7 @@ reg_t *re;
 #  else	/* !USEREGCMP */
 
 reg_t *regexp_init(s, len)
-char *s;
+CONST char *s;
 int len;
 {
 	reg_t *re;
@@ -1308,8 +1379,8 @@ int len;
 }
 
 static int NEAR _regexp_exec(re, s)
-reg_t *re;
-char *s;
+CONST reg_t *re;
+CONST char *s;
 {
 	int i, n1, n2, c1, c2, beg, rev;
 
@@ -1378,8 +1449,8 @@ char *s;
 }
 
 int regexp_exec(re, s, fname)
-reg_t *re;
-char *s;
+CONST reg_t *re;
+CONST char *s;
 int fname;
 {
 	if (!re || (fname && skipdotfile && *s == '.')) return(0);
@@ -1405,7 +1476,7 @@ reg_t *re;
 
 static void NEAR addstrbuf(sp, s, len)
 strbuf_t *sp;
-char *s;
+CONST char *s;
 int len;
 {
 	sp -> s = c_realloc(sp -> s, sp -> len + len, &(sp -> size));
@@ -1415,7 +1486,8 @@ int len;
 }
 
 static void NEAR duplwild(dst, src)
-wild_t *dst, *src;
+wild_t *dst;
+CONST wild_t *src;
 {
 	memcpy((char *)dst, (char *)src, sizeof(wild_t));
 	dst -> fixed.s = (char *)malloc2(src -> fixed.size);
@@ -1448,13 +1520,17 @@ int argc;
 char ***argvp;
 wild_t *wp;
 {
+#ifdef	DOUBLESLASH
+	int ds;
+#endif
 	DIR *dirp;
 	struct dirent *dp;
 	struct stat st;
 	reg_t *re;
 	wild_t dupl;
 	ALLOC_T flen, plen;
-	char *cp;
+	CONST char *cp;
+	char *new;
 	int i, n, w, pc, quote, isdir;
 
 	if (!*(wp -> s)) {
@@ -1471,6 +1547,14 @@ wild_t *wp;
 	quote = wp -> quote;
 
 	if (wp -> fixed.len) addstrbuf(&(wp -> path), rootpath, 1);
+#ifdef	DOUBLESLASH
+	if (wp -> path.len) ds = 0;
+	else if ((ds = isdslash(wp -> s))) {
+		addstrbuf(&(wp -> fixed), wp -> s, ds);
+		addstrbuf(&(wp -> path), wp -> s, ds);
+		wp -> s += ds;
+	}
+#endif
 
 	for (i = w = 0; wp -> s[i] && wp -> s[i] != _SC_; i++) {
 		pc = parsechar(&(wp -> s[i]), -1,
@@ -1517,6 +1601,9 @@ wild_t *wp;
 
 	if (!w) {
 		if (wp -> path.len <= plen) w++;
+#ifdef	DOUBLESLASH
+		else if (ds) st.st_mode = S_IFDIR;
+#endif
 		else if (stat2(wp -> path.s, &st) < 0) return(argc);
 
 		wp -> s += i;
@@ -1527,13 +1614,17 @@ wild_t *wp;
 		}
 
 #ifndef	NODIRLOOP
+# ifdef	DOUBLESLASH
+		if (ds) /*EMPTY*/;
+		else
+# endif
 		if (!w) {
 			wp -> ino = (devino_t *)realloc2(wp -> ino,
 				(wp -> nino + 1) * sizeof(devino_t));
 			wp -> ino[wp -> nino].dev = st.st_dev;
 			wp -> ino[(wp -> nino)++].ino = st.st_ino;
 		}
-#endif
+#endif	/* !NODIRLOOP */
 		return(_evalwild(argc, argvp, wp));
 	}
 
@@ -1550,14 +1641,14 @@ wild_t *wp;
 		re = NULL;
 	}
 	else {
-		cp = malloc2(i + 2);
+		new = malloc2(i + 2);
 		n = 0;
-		if (quote) cp[n++] = quote;
-		memcpy(&(cp[n]), wp -> s, i);
+		if (quote) new[n++] = quote;
+		memcpy(&(new[n]), wp -> s, i);
 		n += i;
-		if (wp -> quote) cp[n++] = wp -> quote;
-		re = regexp_init(cp, n);
-		free(cp);
+		if (wp -> quote) new[n++] = wp -> quote;
+		re = regexp_init(new, n);
+		free(new);
 		if (!re) return(argc);
 		wp -> s += i + 1;
 	}
@@ -1635,7 +1726,7 @@ CONST VOID_P vp2;
 }
 
 char **evalwild(s, flags)
-char *s;
+CONST char *s;
 int flags;
 {
 	wild_t w;
@@ -1670,7 +1761,7 @@ int flags;
 
 #ifndef	_NOUSEHASH
 static int NEAR calchash(s)
-char *s;
+CONST char *s;
 {
 	u_int n;
 	int i;
@@ -1690,7 +1781,8 @@ static VOID NEAR inithash(VOID_A)
 }
 
 static hashlist *NEAR newhash(com, path, cost, next)
-char *com, *path;
+CONST char *com;
+char *path;
 int cost;
 hashlist *next;
 {
@@ -1733,7 +1825,8 @@ hashlist **htable;
 hashlist **duplhash(htable)
 hashlist **htable;
 {
-	hashlist *hp, **nextp, **new;
+	CONST hashlist *hp;
+	hashlist **nextp, **new;
 	int i;
 
 	if (!htable) return(NULL);
@@ -1754,7 +1847,7 @@ hashlist **htable;
 }
 
 static hashlist *NEAR findhash(com, n)
-char *com;
+CONST char *com;
 int n;
 {
 	hashlist *hp;
@@ -1766,7 +1859,7 @@ int n;
 }
 
 static VOID NEAR rmhash(com, n)
-char *com;
+CONST char *com;
 int n;
 {
 	hashlist *hp, *prev;
@@ -1787,7 +1880,7 @@ int n;
 #endif	/* !_NOUSEHASH */
 
 static int NEAR isexecute(path, dirok, exe)
-char *path;
+CONST char *path;
 int dirok, exe;
 {
 	struct stat st;
@@ -1803,23 +1896,24 @@ int dirok, exe;
 
 #if	MSDOS
 static int NEAR extaccess(path, ext, len, exe)
-char *path, *ext;
+char *path;
+CONST char *ext;
 int len, exe;
 {
 	if (ext) {
 		if (isexecute(path, 0, exe) >= 0) {
-			if (!(strpathcmp(ext, "COM"))) return(0);
-			else if (!(strpathcmp(ext, "EXE"))) return(CM_EXE);
-			else if (!(strpathcmp(ext, "BAT"))) return(CM_BATCH);
+			if (!(strpathcmp(ext, EXTCOM))) return(0);
+			else if (!(strpathcmp(ext, EXTEXE))) return(CM_EXE);
+			else if (!(strpathcmp(ext, EXTBAT))) return(CM_BATCH);
 		}
 	}
 	else {
 		path[len++] = '.';
-		strcpy(&(path[len]), "COM");
+		strcpy(&(path[len]), EXTCOM);
 		if (isexecute(path, 0, exe) >= 0) return(CM_ADDEXT);
-		strcpy(&(path[len]), "EXE");
+		strcpy(&(path[len]), EXTEXE);
 		if (isexecute(path, 0, exe) >= 0) return(CM_ADDEXT | CM_EXE);
-		strcpy(&(path[len]), "BAT");
+		strcpy(&(path[len]), EXTBAT);
 		if (isexecute(path, 0, exe) >= 0) return(CM_ADDEXT | CM_BATCH);
 	}
 
@@ -1829,12 +1923,13 @@ int len, exe;
 
 int searchhash(hpp, com, search)
 hashlist **hpp;
-char *com, *search;
+CONST char *com, *search;
 {
 #if	MSDOS
 	char *ext;
 #endif
-	char *cp, *tmp, *next, *path;
+	CONST char *cp, *next;
+	char *tmp, *path;
 	int len, dlen, cost, size, ret;
 #ifndef	_NOUSEHASH
 	hashlist *hp;
@@ -1907,7 +2002,7 @@ char *com, *search;
 #ifndef	_NOUSEHASH
 	recalc = 0;
 #endif
-	if ((next = (search) ? search : getconstvar("PATH"))) {
+	if ((next = (search) ? search : getconstvar(ENVPATH))) {
 		len = strlen(com);
 		size = ret = 0;
 		path = NULL;
@@ -1924,7 +2019,7 @@ char *com, *search;
 			dlen = (next) ? (next++) - cp : strlen(cp);
 			if (!dlen) tmp = NULL;
 			else {
-				tmp = _evalpath(cp, cp + dlen, 0);
+				tmp = _evalpath(cp, &(cp[dlen]), 0);
 				dlen = strlen(tmp);
 			}
 			if (dlen + len + 1 + EXTWIDTH + 1 > size) {
@@ -1961,26 +2056,28 @@ char *com, *search;
 	return(CM_NOTFOUND);
 }
 
+#ifdef	FD
 char *searchexecpath(path, search)
-char *path, *search;
+CONST char *path, *search;
 {
 	hashlist *hp;
 	int type;
 
 	if ((type = searchhash(&hp, path, search)) & CM_NOTFOUND) return(NULL);
 	if (type & CM_FULLPATH) return(strdup2(path));
-#ifdef	_NOUSEHASH
+# ifdef	_NOUSEHASH
 	return((char *)hp);
-#else
+# else
 	return(strdup2(hp -> path));
-#endif
+# endif
 }
+#endif	/* FD */
 
 #if	!defined (FDSH) && !defined (_NOCOMPLETE)
 char *finddupl(target, argc, argv)
-char *target;
+CONST char *target;
 int argc;
-char **argv;
+char *CONST *argv;
 {
 	int i;
 
@@ -1992,7 +2089,7 @@ char **argv;
 
 # ifndef	NOUID
 int completeuser(name, len, argc, argvp, home)
-char *name;
+CONST char *name;
 int len, argc;
 char ***argvp;
 int home;
@@ -2070,7 +2167,7 @@ int home;
 }
 
 int completegroup(name, len, argc, argvp)
-char *name;
+CONST char *name;
 int len, argc;
 char ***argvp;
 {
@@ -2135,9 +2232,10 @@ char ***argvp;
 # endif	/* !NOUID */
 
 static int NEAR completefile(file, len, argc, argvp, dir, dlen, exe)
-char *file;
+CONST char *file;
 int len, argc;
-char ***argvp, *dir;
+char ***argvp;
+CONST char *dir;
 int dlen, exe;
 {
 	DIR *dirp;
@@ -2181,7 +2279,7 @@ int dlen, exe;
 }
 
 static int NEAR completeexe(file, len, argc, argvp)
-char *file;
+CONST char *file;
 int len, argc;
 char ***argvp;
 {
@@ -2191,7 +2289,7 @@ char ***argvp;
 # ifdef	CWDINPATH
 	argc = completefile(file, len, argc, argvp, curpath, 1, 2);
 # endif
-	if (!(next = getconstvar("PATH"))) return(argc);
+	if (!(next = getconstvar(ENVPATH))) return(argc);
 	for (cp = next; cp; cp = next) {
 # if	MSDOS || (defined (FD) && !defined (_NODOSDRIVE))
 		if (_dospath(cp)) next = strchr(&(cp[2]), PATHDELIM);
@@ -2199,7 +2297,7 @@ char ***argvp;
 # endif
 		next = strchr(cp, PATHDELIM);
 		dlen = (next) ? (next++) - cp : strlen(cp);
-		tmp = _evalpath(cp, cp + dlen, 0);
+		tmp = _evalpath(cp, &(cp[dlen]), 0);
 		dlen = strlen(tmp);
 		argc = completefile(file, len, argc, argvp, tmp, dlen, 2);
 		free(tmp);
@@ -2209,7 +2307,7 @@ char ***argvp;
 }
 
 int completepath(path, len, argc, argvp, exe)
-char *path;
+CONST char *path;
 int len, argc;
 char ***argvp;
 int exe;
@@ -2217,21 +2315,34 @@ int exe;
 # if	MSDOS || (defined (FD) && !defined (_NODOSDRIVE))
 	char cwd[4];
 # endif
-	char *file, *dir;
+# ifdef	DOUBLESLASH
+	int n;
+# endif
+	CONST char *dir, *file;
 	int dlen;
 
-	dir = path;
-	dlen = 0;
 # if	MSDOS || (defined (FD) && !defined (_NODOSDRIVE))
-	if (_dospath(path)) {
-		dir += 2;
-		dlen = 2;
-	}
+	if (_dospath(path)) dlen = 2;
+	else
 # endif
+	dlen = 0;
+	dir = &(path[dlen]);
 
 	if ((file = strrdelim(dir, 0))) {
-		dlen += (file == dir) ? 1 : file - dir;
-		return(completefile(file + 1, strlen(file + 1), argc, argvp,
+		if (file == dir) {
+			dlen++;
+			file++;
+		}
+# ifdef	DOUBLESLASH
+		else if ((n = isdslash(path)) && file < &(path[n])) {
+			if (path[n - 1] != _SC_) return(argc);
+			dlen = n;
+			file = &(path[n]);
+			if (*file == _SC_) file++;
+		}
+# endif
+		else dlen += file++ - dir;
+		return(completefile(file, strlen(file), argc, argvp,
 			path, dlen, exe));
 	}
 # ifndef	NOUID
@@ -2252,7 +2363,7 @@ int exe;
 
 char *findcommon(argc, argv)
 int argc;
-char **argv;
+char *CONST *argv;
 {
 	char *common;
 	int i, n;
@@ -2275,7 +2386,8 @@ char **argv;
 #endif	/* !FDSH && !_NOCOMPLETE */
 
 static int NEAR addmeta(s1, s2, quoted)
-char *s1, *s2;
+char *s1;
+CONST char *s2;
 int quoted;
 {
 	int i, j, pc, quote;
@@ -2301,7 +2413,7 @@ int quoted;
 }
 
 char *catvar(argv, delim)
-char *argv[];
+char *CONST *argv;
 int delim;
 {
 	char *cp;
@@ -2322,7 +2434,7 @@ int delim;
 }
 
 int countvar(var)
-char **var;
+char *CONST *var;
 {
 	int i;
 
@@ -2346,7 +2458,7 @@ char **var;
 }
 
 char **duplvar(var, margin)
-char **var;
+char *CONST *var;
 int margin;
 {
 	char **dupl;
@@ -2365,7 +2477,7 @@ int margin;
 }
 
 int parsechar(s, len, spc, flags, qp, pqp)
-char *s;
+CONST char *s;
 int len, spc, flags, *qp, *pqp;
 {
 	if (*s == *qp) {
@@ -2429,7 +2541,7 @@ int len, spc, flags, *qp, *pqp;
 }
 
 static int NEAR skipvar(bufp, eolp, ptrp, qed)
-char **bufp;
+CONST char **bufp;
 int *eolp, *ptrp, qed;
 {
 	int i, mode;
@@ -2521,22 +2633,22 @@ int *eolp, *ptrp, qed;
 
 #ifdef	MINIMUMSHELL
 static int NEAR skipvarvalue(s, ptrp, next, qed, nonl)
-char *s;
+CONST char *s;
 int *ptrp;
-char *next;
+CONST char *next;
 int qed, nonl;
 #else
 static int NEAR skipvarvalue(s, ptrp, next, qed, nonl, nest)
-char *s;
+CONST char *s;
 int *ptrp;
-char *next;
+CONST char *next;
 int qed, nonl, nest;
 #endif
 {
 #ifdef	NESTINGQUOTE
 	int pq;
 #endif
-	char *cp;
+	CONST char *cp;
 	int pc, q, len;
 
 #ifdef	NESTINGQUOTE
@@ -2585,58 +2697,60 @@ int qed, nonl, nest;
 
 #ifndef	MINIMUMSHELL
 static char *NEAR removeword(s, pattern, plen, mode)
-char *s, *pattern;
+CONST char *s, *pattern;
 int plen, mode;
 {
 	reg_t *re;
-	char *cp, *ret, *tmp;
-	int c, len;
+	CONST char *cp;
+	char *ret, *tmp, *new;
+	int c, n, len;
 
 	if (!s || !*s) return(NULL);
 	tmp = strndup2(pattern, plen);
-	pattern = evalarg(tmp, '\0', EA_BACKQ);
+	new = evalarg(tmp, '\0', EA_BACKQ);
 	free(tmp);
-	re = regexp_init(pattern, -1);
-	free(pattern);
+	re = regexp_init(new, -1);
+	free(new);
 	if (!re) return(NULL);
 	ret = NULL;
 	len = strlen(s);
+	n = -1;
 	if ((mode & ~0x80) != '#') {
-		if (mode & 0x80) for (cp = s; cp < s + len; cp++) {
+		if (mode & 0x80) for (cp = s; cp < &(s[len]); cp++) {
 			if (regexp_exec(re, cp, 0)) {
-				ret = cp;
+				n = cp - s;
 				break;
 			}
 		}
-		else for (cp = s + len - 1; cp >= s; cp--) {
+		else for (cp = &(s[len - 1]); cp >= s; cp--) {
 			if (regexp_exec(re, cp, 0)) {
-				ret = cp;
+				n = cp - s;
 				break;
 			}
 		}
-		if (ret) ret = strndup2(s, ret - s);
+		if (n >= 0) ret = strndup2(s, n);
 	}
 	else {
-		tmp = strdup2(s);
+		new = strdup2(s);
 		if (mode & 0x80)
-		for (cp = tmp + len; cp >= tmp; cp--) {
-			*cp = '\0';
-			if (regexp_exec(re, tmp, 0)) {
-				ret = cp;
+		for (tmp = &(new[len]); tmp >= new; tmp--) {
+			*tmp = '\0';
+			if (regexp_exec(re, new, 0)) {
+				n = tmp - new;
 				break;
 			}
 		}
-		else for (cp = tmp + 1; cp <= tmp + len; cp++) {
-			c = *cp;
-			*cp = '\0';
-			if (regexp_exec(re, tmp, 0)) {
-				ret = cp;
+		else for (tmp = &(new[1]); tmp <= &(new[len]); tmp++) {
+			c = *tmp;
+			*tmp = '\0';
+			if (regexp_exec(re, new, 0)) {
+				n = tmp - new;
 				break;
 			}
-			*cp = c;
+			*tmp = c;
 		}
-		if (ret) ret = strdup2(&(s[ret - tmp]));
-		free(tmp);
+		if (n >= 0) ret = strdup2(&(s[n]));
+		free(new);
 	}
 	regexp_free(re);
 
@@ -2644,7 +2758,8 @@ int plen, mode;
 }
 
 static char **NEAR removevar(var, pattern, plen, mode)
-char **var, *pattern;
+char **var;
+CONST char *pattern;
 int plen, mode;
 {
 	char *cp, **new;
@@ -2669,7 +2784,7 @@ int c, quoted;
 #else
 static char *NEAR evalshellparam(c, quoted, pattern, plen, modep)
 int c, quoted;
-char *pattern;
+CONST char *pattern;
 int plen, *modep;
 #endif
 {
@@ -2702,7 +2817,7 @@ int plen, *modep;
 			}
 #ifdef	BASHSTYLE
 	/* bash uses IFS instead of a space as a separator */
-			if ((cp = getconstvar("IFS")) && *cp)
+			if ((cp = getconstvar(ENVIFS)) && *cp)
 				sp = *cp;
 #endif
 
@@ -2736,7 +2851,7 @@ int plen, *modep;
 			sp = ' ';
 #ifdef	BASHSTYLE
 	/* bash uses IFS instead of a space as a separator */
-			if (quoted && (cp = getconstvar("IFS")))
+			if (quoted && (cp = getconstvar(ENVIFS)))
 				sp = (*cp) ? *cp : '\0';
 #endif
 			cp = catvar(&(arglist[1]), sp);
@@ -2780,7 +2895,8 @@ int plen, *modep;
 }
 
 static int NEAR replacevar(arg, cpp, s, len, vlen, mode, quoted)
-char *arg, **cpp;
+CONST char *arg;
+char **cpp;
 int s, len, vlen, mode, quoted;
 {
 	char *val;
@@ -2841,7 +2957,7 @@ int s, len, vlen, mode, quoted;
 static char *NEAR insertarg(buf, ptr, arg, olen, nlen)
 char *buf;
 int ptr;
-char *arg;
+CONST char *arg;
 int olen, nlen;
 {
 	if (nlen <= olen) return(buf);
@@ -2853,22 +2969,23 @@ int olen, nlen;
 static int NEAR evalvar(bufp, ptr, argp, quoted)
 char **bufp;
 int ptr;
-char **argp;
+CONST char **argp;
 int quoted;
 {
 #ifndef	MINIMUMSHELL
 	char tmp[MAXLONGWIDTH + 1];
 #endif
-	char *cp, *new, *arg, *top;
+	CONST char *arg, *top;
+	char *cp, *new;
 	int i, c, n, len, vlen, s, nul, mode;
 
 	new = NULL;
 	arg = *argp;
-	top = *argp + 1;
+	top = &(arg[1]);
 	n = 0;
 	if ((mode = skipvar(&top, &s, &n, quoted)) < 0) return(-1);
 
-	*argp = top + n - 1;
+	*argp = &(top[n - 1]);
 	if (!(len = s)) {
 		(*bufp)[ptr++] = '$';
 		return(ptr);
@@ -2989,7 +3106,7 @@ int quoted;
 #if	defined (FD) && !defined (NOUID)
 uidtable *finduid(uid, name)
 uid_t uid;
-char *name;
+CONST char *name;
 {
 	struct passwd *pwd;
 	int i;
@@ -3037,7 +3154,7 @@ char *name;
 
 gidtable *findgid(gid, name)
 gid_t gid;
-char *name;
+CONST char *name;
 {
 	struct group *grp;
 	int i;
@@ -3078,7 +3195,9 @@ char *name;
 	gidlist = b_realloc(gidlist, maxgid, gidtable);
 	gidlist[maxgid].gid = grp -> gr_gid;
 	gidlist[maxgid].name = strdup2(grp -> gr_name);
+# ifndef	USEGETGROUPS
 	gidlist[maxgid].gr_mem = duplvar(grp -> gr_mem, -1);
+# endif
 	gidlist[maxgid].ismem = 0;
 
 	return(&(gidlist[maxgid++]));
@@ -3087,13 +3206,29 @@ char *name;
 int isgroupmember(gid)
 gid_t gid;
 {
+# ifdef	USEGETGROUPS
+	gid_t *gidset;
+	int n;
+# else
 	uidtable *up;
+# endif
 	gidtable *gp;
 	int i;
 
 	if (!(gp = findgid(gid, NULL))) return(0);
 	if (!(gp -> ismem)) {
 		gp -> ismem++;
+# ifdef	USEGETGROUPS
+		if ((n = getgroups(0, NULL)) > 0) {
+			gidset = (gid_t *)malloc2(n * sizeof(*gidset));
+			n = getgroups(n, gidset);
+			for (i = 0; i < n; i++) if (gidset[i] == gp -> gid) {
+				gp -> ismem++;
+				break;
+			}
+			free(gidset);
+		}
+# else	/* !USEGETGROUPS */
 		if (gp -> gr_mem && (up = finduid(geteuid(), NULL)))
 		for (i = 0; gp -> gr_mem[i]; i++) {
 			if (!strpathcmp(up -> name, gp -> gr_mem[i])) {
@@ -3103,6 +3238,7 @@ gid_t gid;
 		}
 		freevar(gp -> gr_mem);
 		gp -> gr_mem = NULL;
+# endif	/* !USEGETGROUPS */
 	}
 
 	return(gp -> ismem - 1);
@@ -3123,7 +3259,9 @@ VOID freeidlist(VOID_A)
 	if (gidlist) {
 		for (i = 0; i < maxgid; i++) {
 			free(gidlist[i].name);
+#  ifndef	USEGETGROUPS
 			freevar(gidlist[i].gr_mem);
+#  endif
 		}
 		free(gidlist);
 	}
@@ -3166,7 +3304,7 @@ char *gethomedir(VOID_A)
 #endif	/* !NOUID */
 	char *cp;
 
-	if ((cp = getconstvar("HOME"))) return(cp);
+	if ((cp = getconstvar(ENVHOME))) return(cp);
 #ifndef	NOUID
 # ifdef	FD
 	if ((up = finduid(getuid(), NULL))) return(up -> home);
@@ -3189,21 +3327,57 @@ char *gethomedir(VOID_A)
 	return(NULL);
 }
 
+CONST char *getrealpath(path, resolved, cwd)
+CONST char *path;
+char *resolved, *cwd;
+{
+#if	defined (FD) && !defined (_NODOSDRIVE)
+	int drv;
+#endif
+	CONST char *err;
+	char buf[MAXPATHLEN];
+
+	if (cwd) /*EMPTY*/;
+	else if (!Xgetwd(buf)) return(NULL);
+	else cwd = buf;
+
+#if	defined (FD) && !defined (_NODOSDRIVE)
+	if (preparedrv(dospath3(cwd), &drv) < 0) return(cwd);
+#endif
+	if (_chdir2(path) < 0) err = path;
+	else if (!Xgetwd(resolved)) err = NULL;
+	else if (_chdir2(cwd) < 0) {
+#ifdef	FD
+		lostcwd(NULL);
+		err = resolved;
+#else
+		err = cwd;
+#endif
+	}
+	else err = resolved;
+#if	defined (FD) && !defined (_NODOSDRIVE)
+	shutdrv(drv);
+#endif
+
+	return(err);
+}
+
 #ifndef	MINIMUMSHELL
 int evalhome(bufp, ptr, argp)
 char **bufp;
 int ptr;
-char **argp;
+CONST char **argp;
 {
-	char *cp, *top;
+	CONST char *top;
+	char *cp;
 	int len, vlen;
 
-	top = *argp + 1;
+	top = &((*argp)[1]);
 
 	len = ((cp = strdelim(top, 0))) ? (cp - top) : strlen(top);
 	if (!len) cp = gethomedir();
 # ifdef	FD
-	else if (len == strsize("FD") && !strnpathcmp(top, "FD", len))
+	else if (len == strsize(STRFD) && !strnpathcmp(top, STRFD, len))
 		cp = progpath;
 # endif
 	else {
@@ -3241,13 +3415,13 @@ char **argp;
 	if (!cp) {
 		vlen = len;
 		(*bufp)[ptr++] = '~';
-		cp = top;
 	}
 	else {
-		vlen = strlen(cp);
+		top = cp;
+		vlen = strlen(top);
 		*bufp = insertarg(*bufp, ptr, *argp, len + 1, vlen);
 	}
-	strncpy(&((*bufp)[ptr]), cp, vlen);
+	strncpy(&((*bufp)[ptr]), top, vlen);
 	ptr += vlen;
 	*argp += len;
 
@@ -3268,7 +3442,8 @@ int qed, flags;
 #ifdef	NESTINGQUOTE
 	int pq;
 #endif
-	char *cp, *buf, *bbuf;
+	CONST char *cp;
+	char *buf, *bbuf;
 	int i, j, pc, q;
 
 #if	MSDOS && defined (FD) && !defined (_NOUSELFN)
@@ -3312,7 +3487,7 @@ int qed, flags;
 			if (*cp == '`') {
 				bbuf[j] = '\0';
 				buf = replacebackquote(buf, &i,
-					bbuf, strlen(cp + 1));
+					bbuf, strlen(&(cp[1])));
 				j = 0;
 			}
 			else if (!(flags & EA_STRIPQ)) buf[i++] = *cp;
@@ -3401,7 +3576,8 @@ int qed, flags;
 
 int evalifs(argc, argvp, ifs)
 int argc;
-char ***argvp, *ifs;
+char ***argvp;
+CONST char *ifs;
 {
 	char *cp;
 	int i, j, n, pc, quote;
@@ -3513,12 +3689,15 @@ int flags;
 }
 
 char *_evalpath(path, eol, flags)
-char *path, *eol;
+CONST char *path, *eol;
 int flags;
 {
 #if	MSDOS && defined (FD) && !defined (_NOUSELFN)
 	char alias[MAXPATHLEN];
-	int top = -1;
+	int top;
+#endif
+#ifdef	DOUBLESLASH
+	int ds;
 #endif
 	char *cp, *tmp;
 	int i, j, c, pc, size, quote;
@@ -3534,9 +3713,16 @@ int flags;
 	free(cp);
 	cp = tmp;
 
+#if	MSDOS && defined (FD) && !defined (_NOUSELFN)
+	top = -1;
+#endif
+#ifdef	DOUBLESLASH
+	ds = isdslash(cp);
+#endif
 	size = strlen(cp) + 1;
 	tmp = malloc2(size);
 	quote = '\0';
+
 	for (i = j = c = 0; cp[i]; c = cp[i++]) {
 		pc = parsechar(&(cp[i]), -1, '\0', 0, &quote, NULL);
 		if (pc == PC_CLQUOTE) {
@@ -3571,9 +3757,12 @@ int flags;
 			if (!(flags & EA_NOEVALQ)) continue;
 		}
 		else if (pc != PC_NORMAL) /*EMPTY*/;
-		else if (!(flags & EA_NOUNIQDELIM)
-		&& cp[i] == _SC_ && c == _SC_)
-			continue;
+		else if (flags & EA_NOUNIQDELIM) /*EMPTY*/;
+#ifdef	DOUBLESLASH
+		else if (ds && i == 1) /*EMPTY*/;
+#endif
+		else if (cp[i] == _SC_ && c == _SC_) continue;
+
 		tmp[j++] = cp[i];
 	}
 	tmp[j] = '\0';
